from pydantic import BaseModel
from ibm_watsonx_data_integration.cpd_models.connections_model import Connection

class BaseConnection(BaseModel):
    asset_id: str | None = None
    proj_id: str | None = None

    def from_connection(self, connection: Connection):
        """Initializes the underlying DataStage connection object from a generic cpd connection"""

        self.asset_id = connection.metadata.asset_id
        self.proj_id = connection.metadata.project_id
        self.name = connection.name

        alias_to_attr = {}
        for field_name, field_info in self.__class__.model_fields.items():
            if field_info.alias:
                alias_to_attr[field_info.alias] = field_name

        missing_keys = set()

        for k, v in connection.properties.items():
            if k in alias_to_attr:
                setattr(self, alias_to_attr[k], v)
            else:
                # couldn't find a matching attribute on the Conn object to set
                missing_keys.add(k)

        if missing_keys: print(f"Connection attributes could not be set: {missing_keys}")
        return self



# IBM Confidential
# PID 5900-BAF
# Copyright StreamSets Inc., an IBM Company 2025

"""Authentication module."""

import logging
import requests
import time
from abc import ABC, abstractmethod
from ibm_watsonx_data_integration.common.exceptions import InvalidApiKeyError, InvalidUrlError
from urllib.parse import urlparse, urlunparse

logger = logging.getLogger(__name__)


class BaseAuthenticator(ABC):
    """Base Authenticator classes to be inherited by other authenticators."""

    @abstractmethod
    def get_authorization_header(self) -> str:
        """Returns the token formatted to be plugged into the Authorization header of a request.

        Returns:
            A :py:obj:`str` ready to be used as an Authorization header.
        """
        pass


class IAMAuthenticator(BaseAuthenticator):
    """Authenticator class to authenticate using an IBM Cloud IBM API Key.

    Attributes:
        api_key: API key being used to authenticate.
        token_url: URL being used to authenticate against.
        iam_token: The token generated using the api_key and url.
        token_expiry_time: UNIX time in seconds for when the token will expire.
    """

    # give a time buffer for expiration, to have ample time to make a request without the token being expired.
    EXPIRATION_TIME_BUFFER = 5

    def __init__(self, api_key: str, base_auth_url: str = "https://cloud.ibm.com") -> None:
        """Initializes IAM Authenticator.

        Args:
            api_key: The API key to be used for authentication.
            base_auth_url: The base URL of the IBM cloud instance to be used for authentication.

        Raises:
            InvalidApiKeyError: If api_key is not of type str, or is an empty str.
            InvalidUrlError: If base_auth_url is not of type str, or is an empty str.
            requests.exceptions.HTTPError: If there is an error getting a valid token.
        """
        if not isinstance(api_key, str):
            raise InvalidApiKeyError("api_key should be of type str.")
        if not api_key:
            raise InvalidApiKeyError("api_key should not be an empty str.")

        if not isinstance(base_auth_url, str):
            raise InvalidUrlError("base_auth_url should be of type str.")
        if not base_auth_url:
            raise InvalidUrlError("base_auth_url should not be an empty str.")

        self.api_key = api_key

        # base_auth_url looks like: "https://cloud.ibm.com"
        # token url looks like: "https://iam.cloud.ibm.com/identity/token"
        parsed_url = urlparse(url=base_auth_url)
        new_netloc = "iam." + parsed_url.netloc
        new_path = parsed_url.path + "/identity/token"
        self.token_url = urlunparse((parsed_url.scheme, new_netloc, new_path, "", "", ""))

        self.iam_token = None
        self.token_expiry_time = None
        self.request_token()

    def request_token(self) -> None:
        """Request a token from the servers using the API key.

        Raises:
            InvalidApiKeyError: If api_key is invalid.
            InvalidUrlError: If token_url is invalid.
            requests.exceptions.HTTPError: If there is an error getting a valid token.
        """
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = {"apikey": self.api_key, "grant_type": "urn:ibm:params:oauth:grant-type:apikey"}

        try:
            response = requests.post(self.token_url, headers=headers, data=data)
        except requests.exceptions.ConnectionError as e:
            logger.error("IAMAuthenticator incorrect URL. %s", e)
            raise InvalidUrlError("IAMAuthenticator incorrect URL.")
        if response.status_code == 400:
            raise InvalidApiKeyError("IAMAuthenticator api_key is not valid.")
        response.raise_for_status()

        self.iam_token = response.json()["access_token"]
        self.token_expiry_time = response.json()["expiration"]

    def get_token(self) -> str:
        """Get existing token, or request a new one if the current token is expired.

        Returns:
            A :py:obj:`str` containing the current token.
        """
        current_time = int(time.time())
        if self.iam_token is None:
            logger.debug("Getting first token.")
            self.request_token()
        elif (current_time + self.EXPIRATION_TIME_BUFFER) >= self.token_expiry_time:
            logger.debug("Previous token expired, refreshing.")
            self.request_token()

        return self.iam_token

    def get_authorization_header(self) -> str:
        """Returns the token formatted to be plugged into the Authorization header of a request.

        Returns:
            A :py:obj:`str` ready to be used as an Authorization header.
        """
        return f"Bearer {self.get_token()}"



import ibm_watsonx_data_integration.services.datastage.models.flow_json_model as models
from ibm_watsonx_data_integration.services.datastage.models.enums import FIELD
from ibm_watsonx_data_integration.services.datastage.models.schema.field.base_field import BaseField
from typing import TypeVar
T = TypeVar("T", bound="Bit")

class Bit(BaseField):
    def __init__(self, name: str):
        super().__init__(name)
        self._set_model_type("boolean")
        self._set_app_type_code("BOOLEAN")
        self._set_app_odbc_type("BIT")
        self.configuration.metadata.min_length = 6
        self.configuration.metadata.max_length = 6

    def _create_from_model(self, model: models.FieldModelComplex)-> T:
        field = Bit(model.name)
        field.configuration = model
        return field

    def byte_to_skip(self, num_bytes: int):
        return self._byte_to_skip(num_bytes)

    def delimiter(self, delim: FIELD.Delim):
        return self._delim(delim)

    def delimiter_string(self, delimiter_string: str):
        return self._delim_string(delimiter_string)

    def generate_on_output(self):
        return self._generate_on_output()

    def prefix_bytes(self, prefix_bytes: FIELD.Prefix):
        return self._prefix_bytes(prefix_bytes)

    def quote(self, quote_type: FIELD.Quote):
        return self._quote(quote_type)

    def start_position(self, position: int):
        return self._start_position(position)

    def tag_case_value(self, tag_case_value: int):
        return self._tagcase(tag_case_value)

    def byte_order(self, byte_order: FIELD.ByteOrder):
        return self._byte_order(byte_order)

    def c_format(self, c_format: str):
        return self._c_format(c_format)

    def charset(self, charset: FIELD.CharSet):
        return self._charset(charset)

    def data_format(self, data_format: FIELD.DataFormat):
        return self._data_format(data_format)

    def default(self, default: int):
        return self._default(default)

    def field_max_width(self, max_width: int):
        return self._max_width(max_width)

    def field_width(self, width: int):
        return self._width(width)

    def is_link_field(self):
        return self._link_keep()

    def out_format(self, out_format: str):
        return self._out_format(out_format)

    def padchar(self, padchar: FIELD.PadChar):
        return self._padchar(padchar)

    def generate_type(self, generate_type: FIELD.GenerateType):
        return self._generate_type(generate_type)

    def cycle_increment(self, cycle_increment: FIELD.CycleIncrement | int):
        return self._cycle_increment(cycle_increment)

    def cycle_initial_value(self, cycle_initial_value: FIELD.CycleInitialValue | int):
        return self._cycle_initial_value(cycle_initial_value)

    def cycle_limit(self, cycle_limit: FIELD.CycleLimit | int):
        return self._cycle_limit(cycle_limit)

    def random_limit(self, random_limit: FIELD.RandomLimit):
        return self._random_limit(random_limit)

    def random_seed(self, random_seed: FIELD.RandomSeed):
        return self._random_seed(random_seed)

    def random_signed(self):
        return self._random_signed()

    def null_seed(self, seed: int):
        return self._null_seed(seed)

    def percent_null(self, percent_null: int):
        return self._percent_null(percent_null)



OP_MAPPINGS = {
    "db2": "db2",
    "db2cloud": "db2cloud",
    "bigsql": "bigsql",
    "db2iseries": "db2iseries",
    "DB2ConnectorPX": "db2fordatastage",
    "dashdb": "db2warehouse",
    "db2zos": "db2zos",
    "RedshiftPX": "amazon_redshift",
    "postgresql-amazon": "amazon_postgresql",
    "oracle-amazon": "amazonrds_oracle",
    "AmazonS3PX": "amazons3",
    "mysql": "mysql",
    "mysql-amazon": "mysql_amazon",
    "datastax": "datastax",
    "cloudobjectstoragePX": "cloud_object_storage",
    "WebSphereMQConnectorPX": "ibm_mq",
    "informix": "informix",
    "match360": "match360",
    "cognos-analytics": "cognos_analytics",
    "volumes": "storage_volume",
    "mongodb-ibmcloud": "mongodb_ibmcloud",
    "dvm": "dvm",
    "lakehouse": "watsonx_data",
    "dv": "dv",
    "tm1odata": "planning_analytics",
    "PxDataSet": "dataset",
    "PxFileSet": "fileset",
    "PxLookupFileSet": "lookupfileset",
    "PxSequentialFile": "sequentialfile",
    "generics3": "generics3",
    "box": "box",
    "elasticsearch": "elasticsearch",
    "dropbox": "dropbox",
    "minio": "minio",
    "ftp": "ftp",
    "tableau": "tableau",
    "looker": "google_looker",
    "hdfs-apache": "hdfs_apache",
    "PxCFF": "complex_flat_file",
    "PxExternalSource": "external_source",
    "PxExternalTarget": "external_target",
    "JavaStagePX": "java_integration",
    "HBaseConnectorPX": "apache_hbase",
    "KafkaConnectorPX": "apache_kafka",
    "HiveConnectorPX": "apache_hive",
    "cassandra": "cassandra",
    "CassandraConnectorPX": "cassandra_datastage",
    "impala": "impala",
    "derby": "derby",
    "bigqueryPX": "bigquery",
    "GoogleCloudStoragePX": "google_cloud_storage",
    "GooglePubSubPX": "google_pub_sub",
    "azureblobstorage": "azure_blob_storage",
    "cosmos": "azure_cosmos",
    "databricks": "azure_databricks",
    "azurefilestorage": "azure_file_storage",
    "AzureDatalakePX": "azuredatalake",
    "azuresql": "azuresql",
    "sqlserver": "sqlserver",
    "sapbulkextract": "sapbulkextract",
    "sapdeltaextract": "sapdeltaextract",
    "saphana": "saphana",
    "sapidoc": "sapidoc",
    "sybaseiq": "sapiq",
    "sapodata": "sapodata",
    "PxBapi": "sapbapi",
    "sybase": "sybase",
    "mysql-compose": "mysql_compose",
    "mariadb": "mariadb",
    "postgresql": "postgresql",
    "postgresql-ibmcloud": "postgresql_ibmcloud",
    "greenplum": "greenplum",
    "JDBCConnectorPX": "jdbc",
    "ODBCConnectorPX": "odbc",
    "oracle": "oracle",
    "dremio": "dremio",
    "mongodb": "mongodb",
    "exasol": "exasol",
    "presto": "presto",
    "TeradataConnectorPX": "teradata_datastage",
    "teradata": "teradata",
    "netezza": "netezza",
    "NetezzaConnectorPX": "netezza_optimized",
    "OracleConnectorPX": "oracle_datastage",
    "PxRest": "rest",
    "denodo": "denodo",
    "vertica": "vertica",
    "SnowflakeConnectorPX": "snowflake",
    "singlestore": "singlestore",
    "SALESFORCEJCConnectorPX": "salesforceapi",
    "salesforce": "salesforce",
    "PxFunnel": "funnel",
    "PxJoin": "join",
    "PxMerge": "merge",
    "PxCopy": "copy",
    "PxCombineRecords": "combine_records",
    "PxMakeVect": "make_vector",
    "PxMakeSubRec": "make_subrecord",
    "PxLookup": "lookup",
    "PxPivot": "pivot",
    "PxAggregator": "aggregator",
    "PxCompare": "compare",
    "PxFilter": "filter",
    "PxSort": "sort",
    "PxRemDup": "remove_duplicates",
    "PxChangeCapture": "change_capture",
    "PxBLM": "bloom_filter",
    "PxDifference": "difference",
    "PxSwitch": "switch",
    "PxExternalFilter": "external_filter",
    "PxModify": "modify",
    "PxSample": "sample",
    "PxChangeApply": "change_apply",
    "PxColumnExport": "column_export",
    "PxChecksum": "checksum",
    "PxEncode": "encode",
    "PxColumnImport": "column_import",
    "PxCompress": "compress",
    "PxExpand": "expand",
    "PxWriteRangeMap": "write_range_map",
    "PxDecode": "decode",
    "PxPeek": "peek",
    "PxRowGenerator": "row_generator",
    "PxColumnGenerator": "column_generator",
    "PxSurrogateKeyGeneratorN": "surrogate_key_generator",
    "PxWaveGenerator": "wave_generator",
    "PxSplitSubRec": "split_subrecord",
    "PxPromoteSubRec": "promote_subrecord",
    "PxHead": "head",
    "PxTail": "tail",
    "PxSplitVect": "split_vector",
    "CTransformerStage": "transformer",
    "http": "http",
    "azuresynapse": "azuresynapse",
    "postgresql-azure": "azure_postgresql",
    "PxXMLInput": "xml_input",
    "XMLStagePX": "hierarchical_data",
    "PxXMLOutput": "xml_output",
    "MatchFrequency": "match_frequency",
    "WSTransformerPX": "web_service",
    "Survive": "survive",
    "Standardize": "standardize",
    "Investigate": "investigate",
    "PxSCD": "slowly_changing_dimension",
    "PxGeneric": "generic",
    "AddressVerification2": "address_verification",
    "qsaviop": "address_verification",
    "IADataRule": "data_rules",
    "UnduplicateMatch": "one_source_match",
    "ReferenceMatch": "two_source_match",
}



from ibm_watsonx_data_integration.services.datastage.models.enums import AZUREDATALAKE, AZUREDATALAKE_CONNECTION
from os import environ
from tests.integration.datastage.util import AbstractFlowRunTest
from ibm_watsonx_data_integration.cpd_models.project_model import Project


class TestAzureDataLakeConnection(AbstractFlowRunTest):
    def get_flow_name(self):
        return "azuredatalake_peek"

    def get_fc(self, project: Project):
        fc = project.create_flow(name=self.get_flow_name(), environment=None, description="", flow_type="datastage")

        # Stages
        _fastpath_test_1_ = fc.add_stage("Microsoft Azure Data Lake Storage", "fastpath_test_1")
        _fastpath_test_1_.configuration.escape_character = AZUREDATALAKE.EscapeCharacter.double_quote
        _fastpath_test_1_.configuration.file_format = AZUREDATALAKE.FileFormat.delimited
        _fastpath_test_1_.configuration.file_name = "fastpath_test"
        _fastpath_test_1_.configuration.first_line_is_header = True
        _fastpath_test_1_.configuration.infer_as_varchar = False
        _fastpath_test_1_.configuration.quote_character = AZUREDATALAKE.QuoteCharacter.double_quote

        _fastpath_test_1_.configuration.connection.authentication_method = (
            AZUREDATALAKE_CONNECTION.AuthMethod.client_credentials
        )
        _fastpath_test_1_.configuration.connection.client_id = environ["AZUREDATALAKE_CLIENT_ID"]
        _fastpath_test_1_.configuration.connection.client_secret = environ["AZUREDATALAKE_CLIENT_SECRET"]
        _fastpath_test_1_.configuration.connection.proxy = False
        _fastpath_test_1_.configuration.connection.tenant_id = "bacc55be-2e5e-4227-8635-6dbeec84ee1e"
        _fastpath_test_1_.configuration.connection.url = "https://conopsstorage1.dfs.core.windows.net/tm-ds-storage-1"

        _peek_1_ = fc.add_stage("Peek", "Peek_1")

        # Graph
        fastpath_test_1_schema = _fastpath_test_1_.connect_output_to(_peek_1_).set_name("Link_1").create_schema()

        # Schemas
        fastpath_test_1_schema.add_field(
            "VARCHAR",
            "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
        ).nullable().length(1024)
        fastpath_test_1_schema.add_field(
            "VARCHAR",
            "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa__DUPLICATE__1",
        ).nullable().length(1024)

        project.update_flow(fc)

        return fc

    def validate(self, state):
        assert "Completed" in state



import json
from pathlib import Path
from zipfile import ZipInfo
import traceback

import ibm_watsonx_data_integration.services.datastage.models.flow_json_model as models
from ibm_watsonx_data_integration.services.datastage.codegen.code_generator import (
    FlowCodeGenerator,
    # FunctionLibraryCodeGenerator,
    # JavaLibraryCodeGenerator,
    MasterCodeGenerator,
    # MatchSpecificationCodeGenerator,
    # MessageHandlerCodeGenerator,
    # ParamSetCodeGenerator,
)
from ibm_watsonx_data_integration.services.datastage.codegen.dag_generator import (
    ConnectionGenerator,
    DAGGenerator,
    # FunctionLibraryGenerator,
    # JavaLibraryGenerator,
    # MatchSpecificationGenerator,
)
from ibm_watsonx_data_integration.services.datastage.codegen.exporters.util import (
    _check_create_dir,
    _format_code,
    _generate_flow_name,
    _autogenerated_header,
)
from ibm_watsonx_data_integration.services.datastage.codegen.importers import ZipImporter


class FlowFileExporter:
    def __init__(
        self,
        zip_importer: ZipImporter,
        output_path: str | None,
        *,
        offline: bool = False,
        create_job: bool = True,
        run_job: bool = True,
        use_flow_name: bool = True,
        api_key: str = "<TODO: insert your api_key>",
        project_id: str = "<TODO: insert your project_id>",
    ):
        self.zip_importer = zip_importer
        if output_path is not None and output_path.strip() != "":
            self.output_path = str(Path(output_path).absolute())
            self.write_to_output = True
        else:
            self.output_path = ""
            self.write_to_output = False
        self.offline = offline
        self.create_job = create_job
        self.run_job = run_job
        self.use_flow_name = use_flow_name
        self.api_key = api_key
        self.project_id = project_id
        self.job_objs = set()

    # def _create_subflows(
    #     self,
    #     master_gen: MasterCodeGenerator,
    #     parent_dag: DAG,
    #     paramset_objs: list[ParameterSet],
    #     fc: DataStageFlow,
    # ):
    #     replace_nodes: dict[Node, Node] = {}

    #     for node in parent_dag.nodes():
    #         if isinstance(node, SuperNodeRef):
    #             subflow_name = node.name
    #             subflow_json = json.loads(self.zip_importer.find_subflow(subflow_name))
    #             subflow_model = models.Flow(**subflow_json)
    #             subflow_dag_gen = DAGGenerator(subflow_model)
    #             sfc = subflow_dag_gen.generate()
    #             dag = sfc._dag

    #             self._create_subflows(
    #                 master_gen=master_gen,
    #                 parent_dag=dag,
    #                 paramset_objs=paramset_objs,
    #                 fc=fc,
    #             )

    #             sfc_paramsets: list[ParameterSet] = []
    #             for paramset in sfc.parameter_sets:
    #                 for paramset_obj in paramset_objs:
    #                     if paramset_obj.name == paramset.name:
    #                         sfc_paramsets.append(paramset_obj)

    #             subflow = Subflow(dag=dag, name=subflow_name, is_local=False)
    #             super_node = SuperNode(
    #                 parent_dag=parent_dag,
    #                 subflow_dag=subflow.dag,
    #                 entry_nodes=subflow.entry_nodes,
    #                 exit_nodes=subflow.exit_nodes,
    #                 name=subflow_name,
    #                 label=node.label,
    #                 parameter_sets=sfc_paramsets,
    #                 local_parameters=sfc.local_parameters,
    #                 rcp=node.rcp,
    #             )
    #             replace_nodes[node] = super_node

    #     for node in replace_nodes:
    #         parent_dag.replace_node(node, replace_nodes[node])

    def _export_flow(self, master_gen: MasterCodeGenerator, f_content: bytes, f_info: ZipInfo):
        all_code: list[str] = FlowCodeGenerator().generate_setup(self.api_key, self.project_id)

        flow_json = json.loads(f_content)
        try:
            flow_model = models.Flow(**flow_json)
        except Exception:
            try:
                flow_json = flow_json["attachments"]
                flow_model = models.Flow(**flow_json)
            except Exception:
                try:
                    flow_json = flow_json[0]
                    flow_model = models.Flow(**flow_json)
                except Exception as e:
                    raise ValueError(f"Bad flow json: {e}")

        # paramset_objs: list[ParameterSet] = []
        # if flow_model.external_paramsets:
        #     for paramset in flow_model.external_paramsets:
        #         ps_content = self.zip_importer.find_paramset(paramset.name)
        #         ps_data = json.loads(ps_content)["parameter_set"]
        #         param_set = ParameterSet.from_dict(ps_data)
        #         paramset_objs.append(param_set)
        #         ps_code_gen = ParamSetCodeGenerator(param_set, master_gen)
        #         ps_code = ps_code_gen.generate_code()
        #         var_name = master_gen.get_object_var(param_set)
        #         ps_code += f"\nflow.use_paramset({var_name})\n"
        #         formatted = _format_code(ps_code)
        #         all_code.append(f"{formatted}\n\n")

        for conn_info, conn_content in self.zip_importer.connections:
            conn_data = json.loads(conn_content)
            conn_gen = ConnectionGenerator(conn_data)
            conn_model = conn_gen.create_connection_model()
            conn_model
            # conn_code_gen = ConnectionCodeGenerator(conn_model, master_gen)
            # conn_code = conn_code_gen.generate_code()
            # formatted = _format_code(conn_code)
            # all_code.append(f"{formatted}\n\n")

        # for f_info, f_content in self.zip_importer.data_definitions:
        #     json_data = json.loads(f_content)
        #     data_def = DataDefinition.from_dict(json_data)
        #     data_def_code_gen = DataDefinitionCodeGenerator(data_def, master_gen)
        #     data_def_code = data_def_code_gen.generate_code()
        #     formatted = _format_code(data_def_code)
        #     all_code.append(f"{formatted}\n\n")

        # for mh_info, mh_content in self.zip_importer.message_handlers:
        #     mh_data = json.loads(mh_content)
        #     properties = mh_data["entity"]
        #     properties["name"] = mh_data["name"]
        #     properties["description"] = mh_data["description"]
        #     message_handler = MessageHandler(**properties)
        #     mh_code_gen = MessageHandlerCodeGenerator(message_handler, master_gen)
        #     mh_code = mh_code_gen.generate_code()
        #     formatted = _format_code(mh_code)
        #     all_code.append(f"{formatted}\n\n")

        # for jl_info, jl_content in self.zip_importer.java_libraries:
        #     jl_data = json.loads(jl_content)
        #     jl_gen = JavaLibraryGenerator(jl_data)
        #     jl_model = jl_gen.create_java_library_model(self.output_path)
        #     jl_code_gen = JavaLibraryCodeGenerator(jl_model, self.output_path, master_gen)
        #     jl_code = jl_code_gen.generate_code()
        #     formatted = _format_code(jl_code)
        #     all_code.append(f"{formatted}\n\n")

        # for fl_info, fl_content in self.zip_importer.function_libraries:
        #     fl_data = json.loads(fl_content)
        #     fl_gen = FunctionLibraryGenerator(fl_data)
        #     fl_model = fl_gen.create_function_library_model(self.output_path)
        #     fl_code_gen = FunctionLibraryCodeGenerator(fl_model, self.output_path, master_gen)
        #     fl_code = fl_code_gen.generate_code()
        #     formatted = _format_code(fl_code)
        #     all_code.append(f"{formatted}\n\n")

        # for ms_info, ms_content in self.zip_importer.match_specifications:
        #     ms_data = json.loads(ms_content)
        #     ms_gen = MatchSpecificationGenerator(ms_data)
        #     ms_model = ms_gen.create_match_specification_model(self.output_path)
        #     ms_code_gen = MatchSpecificationCodeGenerator(ms_model, self.output_path, master_gen)
        #     ms_code = ms_code_gen.generate_code()
        #     formatted = _format_code(ms_code)
        #     all_code.append(f"{formatted}\n\n")

        dag_gen = DAGGenerator(flow_model)
        fc = dag_gen.generate()

        # self._create_subflows(
        #     master_gen=master_gen,
        #     parent_dag=fc._dag,
        #     paramset_objs=paramset_objs,
        #     fc=fc,
        # )

        # replace_nodes = {}
        # for node in fc._dag.nodes():
        # if isinstance(node, SuperNodeRef):
        #     subflow_name = node.name
        #     subflow_json = json.loads(self.zip_importer.find_subflow(subflow_name))
        #     subflow_model = models.Flow(**subflow_json)
        #     subflow_dag_gen = DAGGenerator(subflow_model)
        #     sfc = subflow_dag_gen.generate()
        #     dag = sfc._dag

        #     # sfc_paramsets: list[ParameterSet] = []
        #     # for paramset in sfc.parameter_sets:
        #     #     for paramset_obj in paramset_objs:
        #     #         if paramset_obj.name == paramset.name:
        #     #             sfc_paramsets.append(paramset_obj)

        #     subflow = Subflow(dag=dag, name=subflow_name, is_local=False)
        #     super_node = SuperNode(
        #         parent_dag=fc._dag,
        #         subflow_dag=subflow.dag,
        #         entry_nodes=subflow.entry_nodes,
        #         exit_nodes=subflow.exit_nodes,
        #         name=subflow_name,
        #         label=node.label,
        #         parameter_sets=sfc_paramsets,
        #         local_parameters=sfc.local_parameters,
        #     )
        #     replace_nodes[node] = super_node

        # elif isinstance(node, BuildStageStage):
        #     build_stage_name = node.configuration.op_name
        #     build_stage_json = json.loads(self.zip_importer.find_build_stage(build_stage_name))
        #     build_stage_asset = BuildStage.from_dict(build_stage_json)
        #     node.build_stage = build_stage_asset

        # elif isinstance(node, WrappedStageStage):
        #     wrapped_stage_name = node.configuration.op_name
        #     wrapped_stage_json = json.loads(self.zip_importer.find_wrapped_stage(wrapped_stage_name))
        #     wrapped_stage_asset = WrappedStage.from_dict(wrapped_stage_json)
        #     node.wrapped_stage = wrapped_stage_asset

        # elif isinstance(node, CustomStageStage):
        #     custom_stage_name = node.configuration.op_name
        #     custom_stage_json = json.loads(self.zip_importer.find_custom_stage(custom_stage_name))
        #     attachment_info, attachment_content = self.zip_importer.find_custom_stage_attachment(f"lib_{custom_stage_name}")
        #     if self.write_to_output:
        #         _check_create_dir(Path(self.output_path) / "attachments")
        #         attachment_path = Path(self.output_path) / "attachments" / (Path(attachment_info.filename).stem + ".so")
        #         with open(attachment_path, "w") as f:
        #             f.write(str(base64.b64encode(attachment_content))[2:-1])
        #     if "entity" in custom_stage_json:
        #         custom_stage_json["entity"]["library_path"] = str(attachment_path)
        #     else:
        #         custom_stage_json["library_path"] = str(attachment_path)
        #     custom_stage_asset = CustomStage.from_dict(custom_stage_json)
        #     node.custom_stage = custom_stage_asset

        # for node in replace_nodes:
        #     fc._dag.replace_node(node, replace_nodes[node])

        if self.use_flow_name:
            flow_name = Path(f_info.filename).stem
        else:
            flow_name = _generate_flow_name()

        code_gen = FlowCodeGenerator(flow_name, fc, master_gen, offline=self.offline, skip_subflows=True)
        code = code_gen.generate_all_zip_one_file()
        all_code.append(code)

        all_code.append(f"\n\nproject.update_flow({code_gen.composer})")

        if self.create_job:
            job_obj = object()
            self.job_objs.add(job_obj)
            job_var = master_gen.reserve_var(f"{flow_name}_job", job_obj)
            all_code.append(f'\n\n{job_var} = project.create_job(name="{flow_name}_job", flow={code_gen.composer})')
            if self.run_job:
                job_run_obj = object()
                self.job_objs.add(job_run_obj)
                job_run_var = master_gen.reserve_var(f"{flow_name}_job_run", job_run_obj)
                all_code.append(f'\n\n{job_run_var} = {job_var}.start(name="{flow_name}_job_run", description="")')

        return _format_code("\n".join(all_code))

    def run(self):
        generated_code = {}
        errors = {}
        if self.write_to_output:
            _check_create_dir(self.output_path)

        master_gen = MasterCodeGenerator()
        for f_info, f_content in self.zip_importer.flows:
            try:
                code = _autogenerated_header() + self._export_flow(master_gen, f_content, f_info)
                file_name = Path(self.output_path) / (Path(f_info.filename).stem + ".py")
                generated_code[str(file_name)] = code
                if self.write_to_output:
                    with open(file_name, "w") as f:
                        f.write(code)
            except Exception:
                errors[Path(f_info.filename).stem] = traceback.format_exc()

        return generated_code, errors



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing utils."""

import logging
import random
import string
import yaml
from functools import wraps
from ibm_watsonx_data_integration.common.constants import DATASTAGE
from requests.exceptions import HTTPError
from time import sleep, time
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.common.models import BaseModel
    from ibm_watsonx_data_integration.platform import Platform

logger = logging.getLogger()


def _get_service_id_to_name_map(platform: "Platform") -> dict:
    """Returns a map of Service ID to Service name.

    Args:
        platform: A Platform object.

    Returns:
        A (:obj:`dict`) containing the Service ID to Service name mapping.
    """
    q_string = ""
    for name in [DATASTAGE]:
        q_string += f"name:{name} "

    catalog = platform._global_catalog_api.get_global_catalog(q_string=q_string, complete=False).json()

    map = {}
    for entity in catalog["resources"]:
        map[entity["_id"]] = entity["name"]

    return map


def get_random_string(characters: str = string.ascii_letters, length: int = 8) -> str:
    """Returns a string of the requested length consisting of random combinations of string characters.

    Args:
        characters:The combination of characters.
        length: Length of String.

    Returns:
        A (:obj:`str`) object.
    """
    return "".join(random.choice(characters) for _ in range(length))


def matches_filters(obj: "BaseModel", **kwargs: dict) -> bool:
    """Returns true if the obj repr args matches the key/value provided from kwargs."""
    repr_args = dict(obj.__repr_args__())
    for key in kwargs:
        if key not in repr_args:
            raise ValueError(f"Cannot filter on {key}")
        if repr_args[key] != kwargs[key]:
            return False
    return True


class SeekableList(list):
    """A list where we retrieve objects based on the matches_filters method."""

    def get(self, **kwargs: dict) -> "BaseModel":
        """Retrieve the first instance that matches the supplied arguments.

        Args:
            **kwargs: Optional arguments to be passed to filter the results offline.

        Returns:
            The first instance from the group that matches the supplied arguments.
        """
        try:
            return next(i for i in self if matches_filters(i, **kwargs))
        except StopIteration:
            raise ValueError("Instance ({}) is not in list".format(", ".join(f"{k}={v}" for k, v in kwargs.items())))

    def get_all(self, **kwargs: dict) -> "SeekableList[BaseModel]":
        """Retrieve all instances that match the supplied arguments.

        Args:
            **kwargs: Optional arguments to be passed to filter the results offline.

        Returns:
            A :py:obj:`ibm_watsonx_data_integration.common.utils.SeekableList` of results
            that match the supplied arguments.
        """
        return SeekableList(i for i in self if matches_filters(i, **kwargs))


class TraversableDict(dict):
    """A dictionary where they key can be a string used to navigate multiple levels."""

    def __init__(self, data: dict, string_key_splitter: str = ".") -> None:
        """ """  # noqa: D419
        super().__init__(data)
        self.string_key_splitter = string_key_splitter

    def __setitem__(self, key_path: str, value: Any) -> None:  # noqa: ANN401
        """ """  # noqa: D419
        keys = key_path.split(self.string_key_splitter)
        if len(keys) <= 1:
            return super().__setitem__(keys[0], value)

        tmp_data = super().__getitem__(keys[0])
        # Traverse to the key depth
        for key in keys[1:-1]:
            if isinstance(tmp_data, list):
                key = int(key)

            tmp_data = tmp_data[key]
        return tmp_data.__setitem__(keys[-1], value)

    def __getitem__(self, key_path: str) -> Any:  # noqa: ANN401
        """ """  # noqa: D419
        # Returns Value given key path
        keys = key_path.split(self.string_key_splitter)
        tmp_data = super().__getitem__(keys[0])
        # Traverse to the key depth
        for key in keys[1:]:
            if isinstance(tmp_data, list):
                key = int(key)

            tmp_data = tmp_data[key]
        return tmp_data

    def __delitem__(self, key_path: str) -> None:
        """ """  # noqa: D419
        keys = key_path.split(self.string_key_splitter)
        if len(keys) <= 1:
            return super().__delitem__(keys[0])

        tmp_data = super().__getitem__(keys[0])
        # Traverse to the key depth
        for key in keys[1:-1]:
            if isinstance(tmp_data, list):
                key = int(key)

            tmp_data = tmp_data[key]
        del tmp_data[keys[-1]]

    def __contains__(self, key_path: str) -> bool:
        """ """  # noqa: D419
        keys = key_path.split(self.string_key_splitter)
        if len(keys) <= 1:
            return super().__contains__(keys[0])
        tmp_data = super().__getitem__(keys[0])
        # Traverse to the key depth
        for key in keys[1:-1]:
            if key in tmp_data:
                if isinstance(tmp_data, list):
                    key = int(key)

                tmp_data = tmp_data[key]
            else:
                return False
        return keys[-1] in tmp_data


def wait_and_retry_on_http_error(timeout_sec: int = 300, time_between_calls_sec: int = 2) -> any:
    """Decorator to retry a function on HTTPError for draft runs and snapshots using tunneling.

    Retries the decorated function upon encountering an HTTPError, logging a warning and waiting
    between attempts until the specified timeout is reached.

    Args:
        func: The function to decorate.
        timeout_sec: Maximum time to keep retrying (default is 300 seconds).
        time_between_calls_sec: Wait time between retry attempts (default is 2 seconds).
    """

    def decorator(func: callable) -> any:
        @wraps(func)
        def inner(*args: any, **kwargs: any) -> any:
            end_time = time() + timeout_sec
            error = None
            while time() < end_time:
                try:
                    return func(*args, **kwargs)
                except HTTPError as e:
                    error = e
                    logger.warning(
                        "HTTPError occurred. Retrying in a bit.",
                        extra=dict(error=e, wait_time_for_next_call=time_between_calls_sec),
                    )
                    sleep(time_between_calls_sec)
            raise error

        return inner

    return decorator


def get_params_from_swagger(content_string: str, request_path: str) -> list:
    """Extract the request params from file.

    Args:
        content_string: The full swagger defintion in string format.
        request_path: The path of the request you want the parameters from.

    Returns:
        A list of request parameters.
    """
    request_params = []
    data = TraversableDict(yaml.safe_load(content_string), "%")
    param_locations = data["paths"][request_path]["get"]["parameters"]
    for param_location in param_locations:
        param_location = param_location["$ref"]
        param_location = param_location.replace("#/", "").replace("/", "%")
        request_params.append(data[param_location]["name"].replace("entity.", "").replace("metadata.", ""))
    return request_params



from os import environ
from tests.integration.datastage.util import AbstractFlowRunTest
from ibm_watsonx_data_integration.cpd_models.project_model import Project


class TestNetezzaDSConnection(AbstractFlowRunTest):
    def get_flow_name(self):
        return "netezzaDS"

    def get_fc(self, project: Project):
        fc = project.create_flow(name=self.get_flow_name(), environment=None, description="", flow_type="datastage")

        # Stages
        _fastpath_jdbc_netezza_1_ = fc.add_stage(
            "IBM Netezza Performance Server for DataStage", "FASTPATH_JDBC_NETEZZA_1"
        )
        _fastpath_jdbc_netezza_1_.configuration.table_name = '"TM_DS_SCHEMA"."FASTPATH_JDBC_NETEZZA"'

        _fastpath_jdbc_netezza_1_.configuration.connection.cas_lite_service_authorization_header = ""
        _fastpath_jdbc_netezza_1_.configuration.connection.database = "TM_DS"
        _fastpath_jdbc_netezza_1_.configuration.connection.hostname = environ["NETEZZA_DS_HOST"]
        _fastpath_jdbc_netezza_1_.configuration.connection.password = environ["NETEZZA_DS_PASSWORD"]
        _fastpath_jdbc_netezza_1_.configuration.connection.ssl_connection = False
        _fastpath_jdbc_netezza_1_.configuration.connection.use_separate_connection_for_twt = False
        _fastpath_jdbc_netezza_1_.configuration.connection.username = environ["NETEZZA_DS_USERNAME"]
        _fastpath_jdbc_netezza_1_.configuration.connection.port = "33262"

        _peek_1_ = fc.add_stage("Peek", "Peek_1")

        # Graph
        fastpath_jdbc_netezza_1_schema = (
            _fastpath_jdbc_netezza_1_.connect_output_to(_peek_1_).set_name("Link_1").create_schema()
        )

        # Schemas
        fastpath_jdbc_netezza_1_schema.add_field("CHAR", "COLUMN_1").length(6)
        fastpath_jdbc_netezza_1_schema.add_field("CHAR", "COLUMN_2").length(6)

        project.update_flow(fc)

        return fc

    def validate(self, state):
        assert "Completed" in state



# -------------------------------------------------------------------
# This file was auto-generated. Do not edit directly.
# Generated by: StageGeneratorV2
# -------------------------------------------------------------------

from .address_verification import AddressVerificationStage
from .aggregator import AggregatorStage
from .mysql_amazon import AmazonRDSForMysqlStage
from .amazonrds_oracle import AmazonRDSForOracleStage
from .amazon_postgresql import AmazonRDSForPostgresqlStage
from .amazon_redshift import AmazonRedshiftStage
from .amazons3 import AmazonS3Stage
from .cassandra import ApacheCassandraStage
from .cassandra_datastage import ApacheCassandraForDatastageStage
from .derby import ApacheDerbyStage
from .hdfs_apache import ApacheHDFSStage
from .apache_hbase import ApacheHbaseStage
from .apache_hive import ApacheHiveStage
from .impala import ApacheImpalaStage
from .apache_kafka import ApacheKafkaStage
from .azure_postgresql import AzurePostgresqlStage
from .bloom_filter import BloomFilterStage
from .box import BoxStage
from .change_apply import ChangeApplyStage
from .change_capture import ChangeCaptureStage
from .checksum import ChecksumStage
from .column_export import ColumnExportStage
from .column_generator import ColumnGeneratorStage
from .column_import import ColumnImportStage
from .combine_records import CombineRecordsStage
from .compare import CompareStage
from .complex_flat_file import ComplexFlatFileStage
from .compress import CompressStage
from .copy import CopyStage
from .data_rules import DataRulesStage
from .dataset import DataSetStage
from .datastax import DatastaxEnterpriseStage
from .decode import DecodeStage
from .denodo import DenodoStage
from .difference import DifferenceStage
from .dremio import DremioStage
from .dropbox import DropboxStage
from .elasticsearch import ElasticsearchStage
from .encode import EncodeStage
from .exasol import ExasolStage
from .expand import ExpandStage
from .external_filter import ExternalFilterStage
from .external_source import ExternalSourceStage
from .external_target import ExternalTargetStage
from .ftp import FTPStage
from .fileset import FileSetStage
from .filter import FilterStage
from .funnel import FunnelStage
from .generic import GenericStage
from .jdbc import GenericJDBCStage
from .generics3 import GenericS3Stage
from .bigquery import GoogleBigqueryStage
from .google_pub_sub import GoogleCloudPubSubStage
from .google_cloud_storage import GoogleCloudStorageStage
from .google_looker import GoogleLookerStage
from .greenplum import GreenplumStage
from .http import HTTPStage
from .head import HeadStage
from .hierarchical_data import HierarchicalDataStage
from .mongodb_ibmcloud import IBMCloudDatabasesForMongodbStage
from .mysql_compose import IBMCloudDatabasesForMysqlStage
from .postgresql_ibmcloud import IBMCloudDatabasesForPostgresqlStage
from .cloud_object_storage import IBMCloudObjectStorageStage
from .cognos_analytics import IBMCognosAnalyticsStage
from .dv import IBMDataVirtualizationStage
from .dvm import IBMDataVirtualizationManagerForZOSStage
from .db2 import IBMDb2Stage
from .bigsql import IBMDb2BigSQLStage
from .db2fordatastage import IBMDb2ForDatastageStage
from .db2iseries import IBMDb2ForIStage
from .db2zos import IBMDb2ForZOSStage
from .db2cloud import IBMDb2OnCloudStage
from .db2warehouse import IBMDb2WarehouseStage
from .informix import IBMInformixStage
from .ibm_mq import IBMMQStage
from .match360 import IBMMatch360Stage
from .netezza import IBMNetezzaPerformanceServerStage
from .netezza_optimized import IBMNetezzaPerformanceServerForDatastageStage
from .planning_analytics import IBMPlanningAnalyticsStage
from .watsonx_data import IBMWatsonxDataPrestoStage
from .investigate import InvestigateStage
from .java_integration import JavaIntegrationStage
from .join import JoinStage
from .lookup import LookupStage
from .lookupfileset import LookupFileSetStage
from .make_subrecord import MakeSubrecordStage
from .make_vector import MakeVectorStage
from .mariadb import MariadbStage
from .match_frequency import MatchFrequencyStage
from .merge import MergeStage
from .azure_blob_storage import MicrosoftAzureBlobStorageStage
from .azure_cosmos import MicrosoftAzureCosmosDBStage
from .azuredatalake import MicrosoftAzureDataLakeStorageStage
from .azure_databricks import MicrosoftAzureDatabricksStage
from .azure_file_storage import MicrosoftAzureFileStorageStage
from .azuresql import MicrosoftAzureSQLDatabaseStage
from .azuresynapse import MicrosoftAzureSynapseAnalyticsStage
from .sqlserver import MicrosoftSQLServerStage
from .minio import MinioStage
from .modify import ModifyStage
from .mongodb import MongodbStage
from .mysql import MysqlStage
from .odbc import ODBCStage
from .one_source_match import OneSourceMatchStage
from .oracle import OracleStage
from .oracle_datastage import OracleDatabaseForDatastageStage
from .peek import PeekStage
from .pivot import PivotEnterpriseStage
from .postgresql import PostgresqlStage
from .presto import PrestoStage
from .promote_subrecord import PromoteSubrecordStage
from .remove_duplicates import RemoveDuplicatesStage
from .rest import RestStage
from .row_generator import RowGeneratorStage
from .sybase import SAPASEStage
from .sapbapi import SAPBAPIStage
from .sapbulkextract import SAPBulkExtractStage
from .sapdeltaextract import SAPDeltaExtractStage
from .saphana import SAPHANAStage
from .sapiq import SAPIQStage
from .sapidoc import SAPIdocStage
from .sapodata import SAPOdataStage
from .salesforceapi import SalesforceAPIForDatastageStage
from .salesforce import SalesforceComStage
from .sample import SampleStage
from .sequentialfile import SequentialFileStage
from .singlestore import SinglestoredbStage
from .slowly_changing_dimension import SlowlyChangingDimensionStage
from .snowflake import SnowflakeStage
from .sort import SortStage
from .split_subrecord import SplitSubrecordStage
from .split_vector import SplitVectorStage
from .standardize import StandardizeStage
from .storage_volume import StorageVolumeStage
from .surrogate_key_generator import SurrogateKeyGeneratorStage
from .survive import SurviveStage
from .switch import SwitchStage
from .tableau import TableauStage
from .tail import TailStage
from .teradata import TeradataStage
from .teradata_datastage import TeradataDatabaseForDatastageStage
from .transformer import TransformerStage
from .two_source_match import TwoSourceMatchStage
from .vertica import VerticaStage
from .wave_generator import WaveGeneratorStage
from .web_service import WebServiceStage
from .write_range_map import WriteRangeMapStage
from .xml_input import XMLInputStage
from .xml_output import XMLOutputStage


__all__ = [
    'AddressVerificationStage',
    'AggregatorStage',
    'AmazonRDSForMysqlStage',
    'AmazonRDSForOracleStage',
    'AmazonRDSForPostgresqlStage',
    'AmazonRedshiftStage',
    'AmazonS3Stage',
    'ApacheCassandraStage',
    'ApacheCassandraForDatastageStage',
    'ApacheDerbyStage',
    'ApacheHDFSStage',
    'ApacheHbaseStage',
    'ApacheHiveStage',
    'ApacheImpalaStage',
    'ApacheKafkaStage',
    'AzurePostgresqlStage',
    'BloomFilterStage',
    'BoxStage',
    'ChangeApplyStage',
    'ChangeCaptureStage',
    'ChecksumStage',
    'ColumnExportStage',
    'ColumnGeneratorStage',
    'ColumnImportStage',
    'CombineRecordsStage',
    'CompareStage',
    'ComplexFlatFileStage',
    'CompressStage',
    'CopyStage',
    'DataRulesStage',
    'DataSetStage',
    'DatastaxEnterpriseStage',
    'DecodeStage',
    'DenodoStage',
    'DifferenceStage',
    'DremioStage',
    'DropboxStage',
    'ElasticsearchStage',
    'EncodeStage',
    'ExasolStage',
    'ExpandStage',
    'ExternalFilterStage',
    'ExternalSourceStage',
    'ExternalTargetStage',
    'FTPStage',
    'FileSetStage',
    'FilterStage',
    'FunnelStage',
    'GenericStage',
    'GenericJDBCStage',
    'GenericS3Stage',
    'GoogleBigqueryStage',
    'GoogleCloudPubSubStage',
    'GoogleCloudStorageStage',
    'GoogleLookerStage',
    'GreenplumStage',
    'HTTPStage',
    'HeadStage',
    'HierarchicalDataStage',
    'IBMCloudDatabasesForMongodbStage',
    'IBMCloudDatabasesForMysqlStage',
    'IBMCloudDatabasesForPostgresqlStage',
    'IBMCloudObjectStorageStage',
    'IBMCognosAnalyticsStage',
    'IBMDataVirtualizationStage',
    'IBMDataVirtualizationManagerForZOSStage',
    'IBMDb2Stage',
    'IBMDb2BigSQLStage',
    'IBMDb2ForDatastageStage',
    'IBMDb2ForIStage',
    'IBMDb2ForZOSStage',
    'IBMDb2OnCloudStage',
    'IBMDb2WarehouseStage',
    'IBMInformixStage',
    'IBMMQStage',
    'IBMMatch360Stage',
    'IBMNetezzaPerformanceServerStage',
    'IBMNetezzaPerformanceServerForDatastageStage',
    'IBMPlanningAnalyticsStage',
    'IBMWatsonxDataPrestoStage',
    'InvestigateStage',
    'JavaIntegrationStage',
    'JoinStage',
    'LookupStage',
    'LookupFileSetStage',
    'MakeSubrecordStage',
    'MakeVectorStage',
    'MariadbStage',
    'MatchFrequencyStage',
    'MergeStage',
    'MicrosoftAzureBlobStorageStage',
    'MicrosoftAzureCosmosDBStage',
    'MicrosoftAzureDataLakeStorageStage',
    'MicrosoftAzureDatabricksStage',
    'MicrosoftAzureFileStorageStage',
    'MicrosoftAzureSQLDatabaseStage',
    'MicrosoftAzureSynapseAnalyticsStage',
    'MicrosoftSQLServerStage',
    'MinioStage',
    'ModifyStage',
    'MongodbStage',
    'MysqlStage',
    'ODBCStage',
    'OneSourceMatchStage',
    'OracleStage',
    'OracleDatabaseForDatastageStage',
    'PeekStage',
    'PivotEnterpriseStage',
    'PostgresqlStage',
    'PrestoStage',
    'PromoteSubrecordStage',
    'RemoveDuplicatesStage',
    'RestStage',
    'RowGeneratorStage',
    'SAPASEStage',
    'SAPBAPIStage',
    'SAPBulkExtractStage',
    'SAPDeltaExtractStage',
    'SAPHANAStage',
    'SAPIQStage',
    'SAPIdocStage',
    'SAPOdataStage',
    'SalesforceAPIForDatastageStage',
    'SalesforceComStage',
    'SampleStage',
    'SequentialFileStage',
    'SinglestoredbStage',
    'SlowlyChangingDimensionStage',
    'SnowflakeStage',
    'SortStage',
    'SplitSubrecordStage',
    'SplitVectorStage',
    'StandardizeStage',
    'StorageVolumeStage',
    'SurrogateKeyGeneratorStage',
    'SurviveStage',
    'SwitchStage',
    'TableauStage',
    'TailStage',
    'TeradataStage',
    'TeradataDatabaseForDatastageStage',
    'TransformerStage',
    'TwoSourceMatchStage',
    'VerticaStage',
    'WaveGeneratorStage',
    'WebServiceStage',
    'WriteRangeMapStage',
    'XMLInputStage',
    'XMLOutputStage',

]



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing Python Generator custom class to hold generated script."""

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


class Code:
    """Utility class to hold generate script of recreating pipeline."""

    def __init__(self, content: str) -> None:
        """The __init__ of the Code class.

        Args:
            content: Generated script as string.
        """
        self._content = content

    def __str__(self) -> str:
        """String representation of containing script."""
        return self._content

    def save(self, path: "Path") -> "Path":
        """Save script content to given path.

        Args:
            path: Location where to save containing script.

        Return:
            Path where generate script was saved.
        """
        with open(path, "w") as f:
            f.write(self._content)

        return path


class Coder(ABC):
    """Interface for classes that should return generated scrip wrapped in ``Code`` class."""

    @abstractmethod
    def to_code(self) -> Code:
        """Here we should return code representation."""



from ibm_watsonx_data_integration.services.datastage.models.enums import (
    AZURE_BLOB_STORAGE,
    AZURE_BLOB_STORAGE_CONNECTION,
)
from ibm_watsonx_data_integration.services.datastage.models.flow.datastage_flow import DataStageFlow
from ibm_watsonx_data_integration.services.datastage.tests.integration.util import AbstractFlowRunTest
from os import environ
from ibm_watsonx_data_integration.cpd_models.project_model import Project


class TestAzureBlobStorageConnection(AbstractFlowRunTest):
    def get_flow_name(self):
        return "azureblobstorage_peek"

    def get_fc(self, project: Project):
        fc = DataStageFlow(project=project, name=self.get_flow_name(), description="", flow_type="datastage")

        # Stages
        _mytestfilecsv_1_ = fc.add_stage("Microsoft Azure Blob Storage", "mytestfilecsv_1")
        _mytestfilecsv_1_.configuration.ds_use_datastage = False
        _mytestfilecsv_1_.configuration.escape_character = AZURE_BLOB_STORAGE.EscapeCharacter.double_quote
        _mytestfilecsv_1_.configuration.file_name = "mytestfile.csv"
        _mytestfilecsv_1_.configuration.infer_as_varchar = False
        _mytestfilecsv_1_.configuration.infer_schema = False
        _mytestfilecsv_1_.configuration.quote_character = AZURE_BLOB_STORAGE.QuoteCharacter.double_quote

        _mytestfilecsv_1_.configuration.connection.authentication_method = (
            AZURE_BLOB_STORAGE_CONNECTION.AuthMethod.connection_string
        )
        _mytestfilecsv_1_.configuration.connection.connection_string = environ["AZUREBLOBSTORAGE_CONN_STRING"]
        _mytestfilecsv_1_.configuration.connection.container = "tm-ds-blob-storage-1"

        _peek_1_ = fc.add_stage("Peek", "Peek_1")

        # Graph
        mytestfilecsv_1_schema = _mytestfilecsv_1_.connect_output_to(_peek_1_).set_name("Link_1").create_schema()

        # Schemas
        mytestfilecsv_1_schema.add_field("VARCHAR", "COLUMN1").nullable().length(1024)
        mytestfilecsv_1_schema.add_field("INTEGER", "COLUMN2").nullable()
        mytestfilecsv_1_schema.add_field("VARCHAR", "COLUMN3").nullable().length(1024)
        mytestfilecsv_1_schema.add_field("DOUBLE", "COLUMN4").nullable()
        mytestfilecsv_1_schema.add_field("INTEGER", "COLUMN5").nullable()
        mytestfilecsv_1_schema.add_field("BIGINT", "COLUMN6").nullable()
        mytestfilecsv_1_schema.add_field("DOUBLE", "COLUMN7").nullable()
        mytestfilecsv_1_schema.add_field("BIT", "COLUMN8").nullable()
        mytestfilecsv_1_schema.add_field("DATE", "COLUMN9").nullable()
        mytestfilecsv_1_schema.add_field("INTEGER", "COLUMN10").nullable()
        mytestfilecsv_1_schema.add_field("INTEGER", "COLUMN11").nullable()
        mytestfilecsv_1_schema.add_field("VARCHAR", "COLUMN12").nullable().length(1024)
        mytestfilecsv_1_schema.add_field("BIGINT", "COLUMN13").nullable()
        mytestfilecsv_1_schema.add_field("INTEGER", "COLUMN14").nullable()
        mytestfilecsv_1_schema.add_field("INTEGER", "COLUMN15").nullable()
        mytestfilecsv_1_schema.add_field("INTEGER", "COLUMN16").nullable()
        mytestfilecsv_1_schema.add_field("DOUBLE", "COLUMN17").nullable()
        mytestfilecsv_1_schema.add_field("INTEGER", "COLUMN18").nullable()
        mytestfilecsv_1_schema.add_field("TIME", "COLUMN19").nullable()
        mytestfilecsv_1_schema.add_field("TIMESTAMP", "COLUMN20").nullable()
        mytestfilecsv_1_schema.add_field("INTEGER", "COLUMN21").nullable()
        mytestfilecsv_1_schema.add_field("BIGINT", "COLUMN22").nullable()
        mytestfilecsv_1_schema.add_field("VARCHAR", "COLUMN23").nullable().length(1024)

        project.update_flow(fc)

        return fc

    def validate(self, state):
        assert "Completed" in state



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing Environment Models."""

import json
import requests
from ibm_watsonx_data_integration.common.json_patch_format import prepare_json_patch_payload
from ibm_watsonx_data_integration.common.models import BaseModel, CollectionModel, CollectionModelResults
from ibm_watsonx_data_integration.services.streamsets.models import Engine
from pydantic import UUID4, Field, PrivateAttr, field_serializer, field_validator
from typing import TYPE_CHECKING, Any, Optional
from urllib.parse import parse_qs, urlparse

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.cpd_models import Project
    from ibm_watsonx_data_integration.platform import Platform


class EnvironmentMetadata(BaseModel):
    """The model for CPD StreamSets Environment Metadata."""

    name: str
    description: str = Field(default="")
    owner_id: str = Field(repr=False, frozen=True)
    create_time: str = Field(repr=False, frozen=True)
    project_id: UUID4 | str = Field(repr=False, frozen=True)
    environment_id: UUID4 | str = Field(alias="asset_id", repr=True, frozen=True)
    asset_attributes: list | None = Field(repr=False, frozen=True)
    asset_state: str | None = Field(repr=False, frozen=True)
    asset_type: str | None = Field(repr=False, frozen=True)
    catalog_id: UUID4 | str | None = Field(repr=False, frozen=True)
    version: float | None = Field(repr=False, frozen=True)

    _expose: bool = PrivateAttr(default=False)

    def model_dump(self, *, by_alias: bool = True, exclude_unset: bool = True, **kwargs: dict) -> dict:
        """Changing default parameters of model_dump to make sure that serialized json math API response.

        Args:
            by_alias: Whether to use alias names in serialization.
            exclude_unset: Whether to exclude unset fields from serialization.
            **kwargs: Additional keyword arguments to pass to the model_dump method.

        Returns:
           A dictionary representation of the model.
        """
        return super().model_dump(exclude_unset=exclude_unset, by_alias=by_alias, **kwargs)


class EngineProperties(BaseModel):
    """The model for engine properties that are included in environments."""

    stagelibs_blocklist: list | None = Field(alias="system.stagelibs.blocklist", default=None)

    @field_validator("stagelibs_blocklist", mode="before")
    def split_string(cls, value: str | None) -> list | None:
        """Splits a stagelibs_blocklist by commas during instance creation."""
        if value is None:
            return None
        return value.split(",")

    @field_serializer("stagelibs_blocklist")
    def serialize_stagelibs(self, stagelibs: list | None) -> str | None:
        """Serializes a list of stage libraries into a comma-separated string (original format)."""
        if stagelibs is None:
            return None
        return ",".join(stagelibs)


class Environment(BaseModel):
    """The model for CPD StreamSets Environment."""

    EXPOSED_DATA_PATH = {"entity.streamsets_environment": {}}
    metadata: EnvironmentMetadata | None = Field(repr=True, default=None)
    engine_type: str = Field(repr=False)
    engine_version: str = Field(repr=False)
    engine_properties: EngineProperties | None = Field(repr=False, default_factory=EngineProperties)
    log4j2_properties: dict[str, str] = Field(repr=False, default_factory=dict)
    external_resource_asset: dict | None = Field(repr=False, default_factory=dict)
    stage_libs: list[str] | None = Field(repr=False, default_factory=list)
    jvm_options: list[str] | None = Field(repr=False, default_factory=list)
    max_cpu_load: float | None = Field(repr=False, default=None)
    max_memory_used: float | None = Field(repr=False, default=None)
    max_jobs_running: int | None = Field(repr=False, default=None)
    cpus_to_allocate: float | None = Field(repr=False, default=None)
    engine_heartbeat_interval: int | None = Field(repr=False, default=None)
    href: str | None = Field(repr=False, exclude=True, default=None)
    engine_statuses: dict[str, str] = Field(repr=False, default_factory=dict)

    def __init__(self, project: "Project", platform: Optional["Platform"] = None, **env_json: dict) -> None:
        """The __init__ of the Environment class.

        Args:
            env_json: The JSON for the Service.
            platform: The Platform object.
            project: The Project object.
        """
        super().__init__(**env_json)
        self._platform = platform
        self._project = project
        self._origin = self.model_dump()

    @staticmethod
    def _create(
        project: "Project",
        name: str,
        engine_version: str | None = None,
        description: str = None,
        engine_type: str = "data_collector",
        engine_properties: dict = None,
        log4j2_properties: dict = None,
        external_resource_asset: dict = None,
        stage_libs: list = None,
        jvm_options: list = None,
        max_cpu_load: float = None,
        max_memory_used: float = None,
        max_jobs_running: int = None,
        engine_heartbeat_interval: int = None,
        cpus_to_allocate: float = None,
    ) -> "Environment":
        if engine_version is None:
            engine_version = project._platform.available_engine_versions[0]

        params = locals()
        params.pop("project")

        env_json = {
            "name": params.pop("name"),
            "description": params.pop("description"),
            "streamsets_environment": {k: v for k, v in params.items() if v is not None},
        }
        query_params = {
            "project_id": project.metadata.guid,
        }
        api_response = project._platform._environment_api.create_environment(  # noqa
            data=json.dumps(env_json), params=query_params
        )
        payload = api_response.json()
        payload.pop("asset_id", None)
        return Environment(project=project, platform=project._platform, **payload)

    def _update(self) -> requests.Response:
        query_params = {
            "project_id": self._project.metadata.guid,
        }
        payload = prepare_json_patch_payload(self.origin, self.model_dump())
        return self._platform._environment_api.patch_environment(  # noqa
            environment_id=self.environment_id, data=payload, params=query_params
        )

    def _delete(self) -> requests.Response:
        query_params = {"project_id": self._project.metadata.guid}
        return self._platform._environment_api.delete_environment(  # noqa
            environment_id=self.environment_id, params=query_params
        )

    @property
    def origin(self) -> dict:
        """Returns origin model dump."""
        return self._origin

    @property
    def environment_id(self) -> str:
        """Returns id of environment."""
        return self.metadata.environment_id

    def get_installation_command(self, pretty: bool = True, foreground: bool = False) -> str | None:
        """Returns the installation command for the environment."""
        query_params = {"project_id": self.metadata.project_id, "pretty": pretty, "foreground": foreground}
        command = self._platform._environment_api.get_docker_run_command(
            environment_id=self.environment_id, params=query_params
        )
        if command.status_code == 200:
            return command.text

    @property
    def engines(self) -> list[Engine]:
        """Returns list of Engines."""
        query_params = {
            "project_id": self.metadata.project_id,
        }
        response = self._platform._environment_api.get_engines(self.environment_id, params=query_params)  # noqa
        return [Engine(platform=self._platform, **data) for data in response.json()["streamsets_engines"]]

    def add_stage_libraries(self, stage_libs: list) -> None:
        """Allows to add stage libraries to the environment.

        Args:
            stage_libs: List of stage libraries to add.
        """
        self.stage_libs.extend(stage_libs)
        self.stage_libs = sorted(set(self.stage_libs))

    def model_dump(self, *, by_alias: bool = True, exclude_unset: bool = True, **kwargs: dict) -> dict:
        """Changing default parameters of model_dump to make sure that serialized json math API response.

        Args:
            by_alias: Whether to use alias names in serialization.
            exclude_unset: Whether to exclude unset fields from serialization.
            **kwargs: Additional keyword arguments to pass to the model_dump method.

        Returns:
            A dictionary representation of the model.
        """
        return super().model_dump(exclude_unset=exclude_unset, by_alias=by_alias, **kwargs)


class Environments(CollectionModel):
    """Collection of Environments instances."""

    def __init__(self, platform: "Platform", project: "Project") -> None:
        """The __init__ of the Environment class.

        Args:
            platform: The Platform object.
            project: The Project object.
        """
        super().__init__(platform=platform)
        self.unique_id = "metadata.environment_id"
        self._project = project

    def __len__(self) -> int:
        """Total amount of environments."""
        params = {"project_id": self._project.metadata.guid, "limit": 0}
        return self._platform._environment_api.get_environments(params=params).json().get("total_count")

    def _request_parameters(self) -> list:
        return ["environment_id", "project_id", "limit", "start", "sort"]

    def _get_results_from_api(self, request_params: dict = None, **kwargs: dict) -> CollectionModelResults:
        """Returns results of an api request."""
        request_params_defaults = {"project_id": self._project.metadata.guid, "limit": 100, "start": None, "sort": None}

        request_params_unioned: dict[str, Any] = request_params_defaults
        request_params_unioned.update(request_params)
        if isinstance(value := request_params_unioned.get("start"), dict):
            request_params_unioned["start"] = parse_qs(urlparse(value.get("href")).query)["start"][0]
        if "environment_id" in request_params:
            response_json = self._platform._environment_api.get_environment(
                environment_id=request_params["environment_id"],
                params={k: v for k, v in request_params_unioned.items() if v is not None},
            ).json()
            response = {"streamsets_environments": [response_json]}
        else:
            response = self._platform._environment_api.get_environments(
                params={k: v for k, v in request_params_unioned.items() if v is not None}
            ).json()
        return CollectionModelResults(
            results=response,
            class_type=Environment,
            response_bookmark="next",
            request_bookmark="start",
            response_location="streamsets_environments",
            constructor_params={"platform": self._platform, "project": self._project},
        )



from enum import Enum
from ibm_watsonx_data_integration.services.datastage.models.connections.base_connection import BaseConnection
from ibm_watsonx_data_integration.services.datastage.models.enums import AZURE_DATABRICKS_CONNECTION
from pydantic import ConfigDict, Field
from typing import ClassVar


class AuthMethod(Enum):
    entra_id = "entra_id"
    oauth_m2m = "oauth_m2m"
    userpass = "userpass"


class AzureDatabricksConn(BaseConnection):
    model_config = ConfigDict(populate_by_name=True)

    datasource_type: ClassVar[str] = "d6cd87ca-ec13-11ee-a951-0242ac120002"
    asset_id: str | None = None
    proj_id: str | None = None
    name: str = Field(None, alias="name")
    authentication_method: AZURE_DATABRICKS_CONNECTION.AuthMethod | None = Field(None, alias="auth_method")
    cluster_access_token: str | None = Field(None, alias="cluster_access_token")
    cluster_user_name: str | None = Field(None, alias="cluster_user_name")
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    microsoft_entra_id_token: str | None = Field(None, alias="entra_id_token")
    hostname_or_ip_address: str = Field(None, alias="host")
    http_path: str | None = Field(None, alias="http_path")
    client_id_of_service_principal: str = Field(None, alias="oauth_m2m_client_id")
    client_secret_of_service_principal: str = Field(None, alias="oauth_m2m_client_secret")
    password: str = Field(None, alias="password")
    port: int = Field(None, alias="port")
    username: str = Field(None, alias="username")
    vaulted_properties: str | None = Field(None, alias="vaulted_properties")
    hidden_dummy_property1: str | None = Field(None, alias="hiddenDummyProperty1")
    hidden_dummy_property2: str | None = Field(None, alias="hiddenDummyProperty2")

    def validate(self):
        include = set()
        exclude = set()

        (
            include.add("microsoft_entra_id_token")
            if (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "entra_id")
                    or (self.authentication_method == "entra_id")
                )
            )
            else exclude.add("microsoft_entra_id_token")
        )
        (
            include.add("password")
            if (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "userpass")
                    or (self.authentication_method == "userpass")
                )
            )
            else exclude.add("password")
        )
        (
            include.add("client_id_of_service_principal")
            if (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "oauth_m2m")
                    or (self.authentication_method == "oauth_m2m")
                )
            )
            else exclude.add("client_id_of_service_principal")
        )
        (
            include.add("client_secret_of_service_principal")
            if (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "oauth_m2m")
                    or (self.authentication_method == "oauth_m2m")
                )
            )
            else exclude.add("client_secret_of_service_principal")
        )
        (
            include.add("username")
            if (
                self.authentication_method
                and (
                    (hasattr(self.authentication_method, "value") and self.authentication_method.value == "userpass")
                    or (self.authentication_method == "userpass")
                )
            )
            else exclude.add("username")
        )
        (
            include.add("hidden_dummy_property2")
            if (self.hidden_dummy_property1)
            else exclude.add("hidden_dummy_property2")
        )
        (
            include.add("hidden_dummy_property1")
            if (self.hidden_dummy_property2)
            else exclude.add("hidden_dummy_property1")
        )
        include.add("vaulted_properties") if (self.hidden_dummy_property1) else exclude.add("vaulted_properties")

        (
            include.add("client_id_of_service_principal")
            if (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "oauth_m2m" in str(self.authentication_method.value)
                    )
                    or ("oauth_m2m" in str(self.authentication_method))
                )
            )
            else exclude.add("client_id_of_service_principal")
        )
        (
            include.add("password")
            if (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "userpass" in str(self.authentication_method.value)
                    )
                    or ("userpass" in str(self.authentication_method))
                )
            )
            else exclude.add("password")
        )
        (
            include.add("username")
            if (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "userpass" in str(self.authentication_method.value)
                    )
                    or ("userpass" in str(self.authentication_method))
                )
            )
            else exclude.add("username")
        )
        (
            include.add("microsoft_entra_id_token")
            if (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "entra_id" in str(self.authentication_method.value)
                    )
                    or ("entra_id" in str(self.authentication_method))
                )
            )
            else exclude.add("microsoft_entra_id_token")
        )
        (
            include.add("client_secret_of_service_principal")
            if (
                self.authentication_method
                and (
                    (
                        hasattr(self.authentication_method, "value")
                        and self.authentication_method.value
                        and "oauth_m2m" in str(self.authentication_method.value)
                    )
                    or ("oauth_m2m" in str(self.authentication_method))
                )
            )
            else exclude.add("client_secret_of_service_principal")
        )
        return include, exclude

    @classmethod
    def from_dict(cls, properties):
        return cls.model_construct(**properties)



# IBM Confidential
# PID 5900-BAF
# Copyright StreamSets Inc., an IBM Company 2025

"""Module with HTTP adapters used for API."""

import requests
from ibm_watsonx_data_integration.__version__ import __version__
from typing import TYPE_CHECKING, Any, Optional

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.common.auth import BaseAuthenticator


class DefaultHTTPAdapter(requests.adapters.HTTPAdapter):
    """Default HTTP Adapter class.

    Every Adapter class used in BaseAPIClient should be a subclass of this one.
    """

    def __init__(
        self,
        pool_connections: int = 1,
        pool_maxsize: int = 10,
        max_retries: int = 0,
        pool_block: bool = False,
        timeout: tuple | float | None = None,
        auth: Optional["BaseAuthenticator"] = None,
    ) -> None:
        """Initialize Adapters.

        https://requests.readthedocs.io/en/latest/api/#requests.adapters.HTTPAdapter

        Args:
            pool_connections: See pool_connections parameter for requests.adapters.HTTPAdapter.
            pool_maxsize: See pool_maxsize parameter for requests.adapters.HTTPAdapter.
            max_retries: See max_retries parameter for requests.adapters.HTTPAdapter.
            pool_block: See pool_block parameter for requests.adapters.HTTPAdapter.
            timeout: See timeout parameter for requests.adapters.HTTPAdapter.send().
            auth: The Authentication object.
        """
        super().__init__(
            pool_connections=pool_connections, pool_maxsize=pool_maxsize, max_retries=max_retries, pool_block=pool_block
        )
        self._auth = auth
        self._timeout = timeout
        self._custom_headers = {}

    def send(self, request: requests.PreparedRequest, **kwargs: dict[str, Any]) -> requests.Response:
        """Overwritten HTTPAdapter.send method.

        https://requests.readthedocs.io/en/latest/api/#requests.adapters.HTTPAdapter.send
        """
        if self._timeout:
            kwargs["timeout"] = self._timeout
        return super().send(request, **kwargs)

    def add_headers(self, request: requests.PreparedRequest, **kwargs: dict[str, Any]) -> None:
        """Overwritten HTTPAdapter.add_headers.

        https://requests.readthedocs.io/en/latest/api/#requests.adapters.HTTPAdapter.add_headers
        """
        # authorization
        if self._auth is not None:
            request.headers["Authorization"] = self._auth.get_authorization_header()

        # required headers
        request.headers.setdefault("Accept", "application/json")
        request.headers.setdefault("Accept-Language", "en-US")
        request.headers.setdefault("Content-Type", "application/json")
        request.headers["watsonx-sdk-version"] = __version__

        request.headers.update(self._custom_headers)



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing AccessGroup Model."""

import json
import requests
from ibm_watsonx_data_integration.common.exceptions import IbmCloudApiException
from ibm_watsonx_data_integration.common.models import BaseModel, CollectionModel, CollectionModelResults
from ibm_watsonx_data_integration.cpd_models.service_id_model import ServiceID
from ibm_watsonx_data_integration.cpd_models.trusted_profile_model import TrustedProfile
from ibm_watsonx_data_integration.cpd_models.user_model import UserProfile
from pydantic import Field
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.platform import Platform


class AccessGroup(BaseModel):
    """Model representing an Access Group, including Rules."""

    id: str = Field(frozen=True, repr=False)
    name: str = Field(repr=True)
    description: str = Field(repr=True)
    created_at: str = Field(repr=False, frozen=True)
    created_by_id: str = Field(repr=False, frozen=True)
    last_modified_at: str = Field(repr=False, frozen=True)
    last_modified_by_id: str = Field(repr=False, frozen=True)

    def __init__(self, platform: Optional["Platform"] = None, **access_group_json: dict) -> None:
        """The __init__ of the AccessGroup class.

        Args:
            platform: The Platform object.
            access_group_json: The JSON for the AccessGroup.
        """
        super().__init__(**access_group_json)
        self._platform = platform

    def __repr__(self) -> str:
        """The access group's representation."""
        desc = self.description
        try:
            self.description = (desc[:75] + "...") if len(desc) > 75 else desc
            return super().__repr__()
        finally:
            self.description = desc

    @staticmethod
    def _create(platform: "Platform", name: str, description: str | None = None) -> "AccessGroup":
        data = {"name": name, "description": description}
        data = json.dumps(data)

        params = {"account_id": platform.current_account.account_id}

        response = platform._access_group_api.create_access_group(params=params, data=data)

        access_group_json = response.json()
        access_group_json["etag"] = response.headers["Etag"]
        return AccessGroup(platform=platform, **access_group_json)

    def _update(self) -> requests.Response:
        data = {"name": self.name, "description": self.description}
        data = json.dumps(data)

        get_ag = self._platform._access_group_api.get_access_group(access_group_id=self.id)
        etag = get_ag.headers["Etag"]

        response = self._platform._access_group_api.update_access_group(access_group_id=self.id, etag=etag, data=data)

        return response

    def _delete(self) -> requests.Response:
        return self._platform._access_group_api.delete_access_group(self.id)

    def add_members_to_access_group(
        self, members: UserProfile | ServiceID | TrustedProfile | list[UserProfile | ServiceID | TrustedProfile]
    ) -> requests.Response:
        """Adds members to an AccessGroup.

        Args:
            members: The members to be added to the access group.

        Returns:
            A HTTP response.
        """
        if not isinstance(members, list):
            m_list = [members]
            members = m_list

        members_list = []
        for m in members:
            members_list.append({"iam_id": m.iam_id, "type": m.type})
        data = {"members": members_list}
        data = json.dumps(data)
        return self._platform._access_group_api.add_members(self.id, data)

    def get_access_group_members(self) -> list:
        """Gets all members of an AccessGroup.

        Returns:
            A list of current members of an Access Group
        """
        ag_members_json_list = self._platform._access_group_api.get_members(self.id).json()["members"]

        user_id_set = set()
        service_id_set = set()
        trusted_profile_id_set = set()

        for member in ag_members_json_list:
            iam_id = member["iam_id"]
            if member["type"] == "user":
                user_id_set.add(iam_id)
            elif member["type"] == "service":
                service_id_set.add(iam_id)
            elif member["type"] == "profile":
                trusted_profile_id_set.add(iam_id)

        users_list = self.get_users(user_id_set)
        service_id_list = self.get_service_ids(service_id_set)
        trusted_profile_id_list = self.get_trusted_profiles(trusted_profile_id_set)

        return [*users_list, *service_id_list, *trusted_profile_id_list]

    def remove_members_from_access_group(
        self, members: UserProfile | ServiceID | TrustedProfile | list[UserProfile | ServiceID | TrustedProfile]
    ) -> requests.Response:
        """Removes members from access group.

        Args:
            members: List of members to remove from access group.

        Returns:
            A HTTP response.
        """
        if not isinstance(members, list):
            m_list = [members]
            members = m_list

        member_ids = []
        for m in members:
            member_ids.append(m.iam_id)

        data = {"members": member_ids}
        data = json.dumps(data)

        return self._platform._access_group_api.remove_members_from_access_group(self.id, data)

    def check_membership(self, member: UserProfile | ServiceID | TrustedProfile) -> requests.Response:
        """Checks the membership of a member in an access group.

        Args:
            member: The member object to check membership of.

        Returns:
            A HTTP response.
        """
        try:
            response = self._platform._access_group_api.check_membership(access_group_id=self.id, iam_id=member.iam_id)
            return response
        except IbmCloudApiException as e:
            return e.response

    def get_users(self, user_id_set: set) -> list:
        """Gets all users with membership to the current access group.

        Args:
            user_id_set: The set of all users in the current account

        Returns:
            list_of_users: List of all users with membership to current access group.
        """
        list_of_users = []

        for user in self._platform.users:
            if user.iam_id in user_id_set:
                list_of_users.append(user)
        return list_of_users

    def get_service_ids(self, service_id_set: set) -> list:
        """Gets all service IDs with membership to the current access group.

        Args:
            service_id_set: The set of all service IDs in the current account

        Returns:
            list_of_service_ids: List of all service IDs with membership to current access group.
        """
        list_of_service_ids = []

        for service_id in self._platform.service_ids:
            if service_id.iam_id in service_id_set:
                list_of_service_ids.append(service_id)

        return list_of_service_ids

    def get_trusted_profiles(self, trusted_profile_set: set) -> list:
        """Gets all trusted profiles with membership to the current access group.

        Args:
            trusted_profile_set: The set of all trusted profiles in the current account

        Returns:
            list_of_trusted_profles: List of all trusted profiles with membership to current access group.
        """
        list_of_trusted_profiles = []

        for trusted_profile in self._platform.trusted_profiles:
            if trusted_profile.iam_id in trusted_profile_set:
                list_of_trusted_profiles.append(trusted_profile)

        return list_of_trusted_profiles


class AccessGroups(CollectionModel):
    """Collection of AccessGroup instances."""

    def __init__(self, platform: Optional["Platform"] = None) -> None:
        """The __init__ of the AccessGroups class.

        Args:
            platform: The Platform object.
        """
        super().__init__(platform)
        self.unique_id = "id"

    def __len__(self) -> int:
        """Total amount of access groups."""
        self._platform.current_account
        return self._platform._access_group_api.get_all_access_groups(
            params={"account_id": self._platform._current_account.account_id}
        ).json()["total_count"]

    def _request_parameters(self) -> list:
        return [
            "account_id",
            "iam_id",
            "search",
            "membership_type",
            "limit",
            "offset",
            "sort",
            "show_federated",
            "hide_public_access",
            "show_crn",
            "access_group_id",
        ]

    def _get_results_from_api(self, request_params: dict = None, **kwargs: dict) -> CollectionModelResults:
        """Returns results of an api request."""
        if "account_id" in request_params:
            next_url = request_params.pop("account_id")
            account_id = next_url["href"].split("account_id=")[1]
            offset = next_url["href"].split("offset=")[1].split("&")[0]
        else:
            account_id = self._platform._current_account.account_id
            offset = 0
        request_params_defaults = {
            "account_id": account_id,
            "iam_id": None,
            "access_group_id": None,
            "search": None,
            "membership_type": None,
            "limit": 100,
            "offset": offset,
            "sort": None,
            "show_federated": None,
            "hide_public_access": None,
            "show_crn": None,
        }
        request_params_unioned = request_params_defaults
        request_params_unioned.update(request_params)

        access_group_id = request_params_unioned.get("access_group_id")

        if access_group_id:
            response = self._platform._access_group_api.get_access_group(access_group_id=access_group_id)
            response = {"groups": [response.json()]}
        else:
            response = self._platform._access_group_api.get_all_access_groups(
                params={k: v for k, v in request_params_unioned.items() if v is not None}
            ).json()
        return CollectionModelResults(
            response,
            AccessGroup,
            "next",
            "account_id",
            "groups",
            {"platform": self._platform},
        )



from os import environ
from tests.integration.datastage.util import AbstractFlowRunTest
from ibm_watsonx_data_integration.cpd_models.project_model import Project


class TestNetezzaConnection(AbstractFlowRunTest):
    def get_flow_name(self):
        return "netezza_peek"

    def get_fc(self, project: Project):
        fc = project.create_flow(name=self.get_flow_name(), environment=None, description="", flow_type="datastage")

        # Stages
        _fastpath_jdbc_netezza_1_ = fc.add_stage("IBM Netezza Performance Server", "FASTPATH_JDBC_NETEZZA_1")
        _fastpath_jdbc_netezza_1_.configuration.schema_name = "TM_DS_SCHEMA"
        _fastpath_jdbc_netezza_1_.configuration.table_name = "FASTPATH_JDBC_NETEZZA"

        _fastpath_jdbc_netezza_1_.configuration.connection.database = "TM_DS"
        _fastpath_jdbc_netezza_1_.configuration.connection.hostname_or_ip_address = environ["NETEZZA_HOST"]
        _fastpath_jdbc_netezza_1_.configuration.connection.password = environ["NETEZZA_PASSWORD"]
        _fastpath_jdbc_netezza_1_.configuration.connection.port = "33262"
        _fastpath_jdbc_netezza_1_.configuration.connection.port_is_ssl_enabled = False
        _fastpath_jdbc_netezza_1_.configuration.connection.username = environ["NETEZZA_USERNAME"]

        _peek_1_ = fc.add_stage("Peek", "Peek_1")

        # Graph
        schema_1 = _fastpath_jdbc_netezza_1_.connect_output_to(_peek_1_).set_name("Link_1").create_schema()

        # Schemas
        schema_1.add_field("CHAR", "COLUMN_1").length(6)
        schema_1.add_field("CHAR", "COLUMN_2").length(6)

        project.update_flow(fc)

        return fc

    def validate(self, state):
        assert "Completed" in state



import warnings
from ibm_watsonx_data_integration.services.datastage.models.base_stage import BaseStage
from ibm_watsonx_data_integration.services.datastage.models.connections.amazon_postgresql_connection import (
    AmazonPostgresqlConn,
)
from ibm_watsonx_data_integration.services.datastage.models.enums import AMAZON_POSTGRESQL
from pydantic import Field
from typing import ClassVar


class amazon_postgresql(BaseStage):

    op_name: ClassVar[str] = "postgresql-amazon"
    node_type: ClassVar[str] = "binding"
    image: ClassVar[str] = "/data-intg/flows/graphics/palette/postgresql-amazon.svg"
    label: ClassVar[str] = "Amazon RDS for PostgreSQL"
    runtime_column_propagation: bool = Field(False, alias="runtime_column_propagation")
    connection: AmazonPostgresqlConn = AmazonPostgresqlConn()
    add_proccode_column: bool | None = Field(False, alias="add_proccode_column")
    batch_size: int | None = Field(2000, alias="batch_size")
    before_after_after_node_fail_on_error: bool | None = Field(True, alias="before_after.after_node.fail_on_error")
    byte_limit: str | None = Field(None, alias="byte_limit")
    call_each_row: bool | None = Field(True, alias="call_each_row")
    call_statement: str | None = Field(None, alias="call_statement")
    create_statement: str | None = Field(None, alias="create_statement")
    default_max_string_binary_precision: int | None = Field(20000, alias="default_max_string_binary_precision")
    defer_credentials: bool | None = Field(False, alias="defer_credentials")
    ds_java_heap_size: int | None = Field(256, alias="_java._heap_size")
    enable_after_sql: str | None = Field("", alias="before_after.after")
    enable_after_sql_node: str | None = Field("", alias="before_after.after_node")
    enable_before_sql: str | None = Field("", alias="before_after.before")
    enable_before_sql_node: str | None = Field("", alias="before_after.before_node")
    execmode: AMAZON_POSTGRESQL.Execmode | None = Field(AMAZON_POSTGRESQL.Execmode.default_par, alias="execmode")
    existing_table_action: AMAZON_POSTGRESQL.ExistingTableAction | None = Field(
        AMAZON_POSTGRESQL.ExistingTableAction.append, alias="existing_table_action"
    )
    fail_on_error_after_sql: bool | None = Field(True, alias="before_after.after.fail_on_error")
    fail_on_error_before_sql: bool | None = Field(True, alias="before_after.before.fail_on_error")
    fail_on_error_before_sql_node: bool | None = Field(True, alias="before_after.before_node.fail_on_error")
    forward_row_data: bool | None = Field(False, alias="forward_row_data")
    generate_unicode_columns: bool | None = Field(False, alias="generate_unicode_columns")
    has_ref_output: bool | None = Field(False, alias="has_ref_output")
    has_reject_output: bool | None = Field(False, alias="has_reject_output")
    input_count: int | None = Field(0, alias="input_count")
    key_column_names: str | None = Field(None, alias="key_column_names")
    login_timeout: int | None = Field(None, alias="login_timeout")
    lookup_type: AMAZON_POSTGRESQL.LookupType | None = Field(AMAZON_POSTGRESQL.LookupType.empty, alias="lookup_type")
    output_count: int | None = Field(0, alias="output_count")
    preserve: AMAZON_POSTGRESQL.Preserve | None = Field(
        AMAZON_POSTGRESQL.Preserve.default_propagate, alias="preserve"
    )
    procedure_name: str | None = Field(None, alias="procedure_name")
    push_filters: str | None = Field(None, alias="push_filters")
    pushed_filters: str | None = Field(None, alias="pushed_filters")
    rcp: bool | None = Field(True, alias="rcp")
    read_after_sql_node_statements_from_file: bool | None = Field(
        False, alias="before_after.after_node.read_from_file_after_sql_node"
    )
    read_after_sql_statements_from_file: bool | None = Field(
        False, alias="before_after.after.read_from_file_after_sql"
    )
    read_before_sql_node_statement_from_file: bool | None = Field(
        False, alias="before_after.before_node.read_from_file_before_sql_node"
    )
    read_before_sql_statements_from_file: bool | None = Field(
        False, alias="before_after.before.read_from_file_before_sql"
    )
    read_mode: AMAZON_POSTGRESQL.ReadMode | None = Field(AMAZON_POSTGRESQL.ReadMode.general, alias="read_mode")
    reject_condition_row_is_rejected: bool | None = Field(False, alias="reject_condition_row_is_rejected")
    reject_data_element_errorcode: bool | None = Field(False, alias="reject_data_element_errorcode")
    reject_data_element_errortext: bool | None = Field(False, alias="reject_data_element_errortext")
    reject_from_link: int | None = Field(None, alias="reject_from_link")
    reject_number: int | None = Field(None, alias="reject_number")
    reject_threshold: int | None = Field(None, alias="reject_threshold")
    reject_uses: AMAZON_POSTGRESQL.RejectUses | None = Field(AMAZON_POSTGRESQL.RejectUses.rows, alias="reject_uses")
    rejected_filters: str | None = Field(None, alias="rejected_filters")
    row_limit: int | None = Field(None, alias="row_limit")
    sampling_percentage: str | None = Field(None, alias="sampling_percentage")
    sampling_seed: int | None = Field(None, alias="sampling_seed")
    sampling_type: AMAZON_POSTGRESQL.SamplingType | None = Field(
        AMAZON_POSTGRESQL.SamplingType.none, alias="sampling_type"
    )
    schema_name: str | None = Field(None, alias="schema_name")
    select_statement: str = Field(None, alias="select_statement")
    select_statement_read_from_file_select: bool | None = Field(
        False, alias="select_statement.read_from_file_select"
    )
    static_statement: str = Field(None, alias="static_statement")
    table_action: AMAZON_POSTGRESQL.TableAction | None = Field(
        AMAZON_POSTGRESQL.TableAction.append, alias="table_action"
    )
    table_name: str = Field(None, alias="table_name")
    transform: str | None = Field("false", alias="transform")
    update_statement: str | None = Field(None, alias="update_statement")
    update_statement_read_from_file_update: bool | None = Field(
        False, alias="update_statement.read_from_file_update"
    )
    user_defined_function: bool | None = Field(None, alias="user_defined_function")
    write_mode: AMAZON_POSTGRESQL.WriteMode | None = Field(AMAZON_POSTGRESQL.WriteMode.insert, alias="write_mode")

    def validate_parameters(self) -> tuple[set, set]:
        include = set()
        exclude = set()

        include.add("preserve") if (self.output_count and self.output_count > 0) else exclude.add("preserve")
        (
            include.add("reject_uses")
            if (self.reject_from_link and self.reject_from_link > -1)
            else exclude.add("reject_uses")
        )
        (
            include.add("reject_number")
            if (self.reject_from_link and self.reject_from_link > -1)
            else exclude.add("reject_number")
        )
        (
            include.add("reject_data_element_errorcode")
            if (self.reject_from_link and self.reject_from_link > -1)
            else exclude.add("reject_data_element_errorcode")
        )
        (
            include.add("reject_data_element_errortext")
            if (self.reject_from_link and self.reject_from_link > -1)
            else exclude.add("reject_data_element_errortext")
        )
        include.add("reject_threshold") if (self.reject_uses == "percent") else exclude.add("reject_threshold")
        return include, exclude

    def validate_source(self) -> tuple[set, set]:
        include = set()
        exclude = set()

        (
            include.add("call_statement")
            if (
                ((not self.table_name))
                and (not self.select_statement)
                and (not self.procedure_name)
                and (

                        self.read_mode
                        and (
                            (hasattr(self.read_mode, "value") and self.read_mode.value == "call_statement")
                            or (self.read_mode == "call_statement")
                        )

                )
            )
            else exclude.add("call_statement")
        )
        (
            include.add("row_limit")
            if (

                    (
                        (
                            self.read_mode
                            and (
                                (hasattr(self.read_mode, "value") and self.read_mode.value != "call")
                                or (self.read_mode != "call")
                            )
                        )
                    )
                    and (

                            self.read_mode
                            and (
                                (hasattr(self.read_mode, "value") and self.read_mode.value != "call_statement")
                                or (self.read_mode != "call_statement")
                            )

                    )

            )
            else exclude.add("row_limit")
        )
        (
            include.add("select_statement")
            if (
                ((not self.schema_name))
                and (not self.table_name)
                and (not self.procedure_name)
                and (not self.call_statement)
                and (

                        self.read_mode
                        and (
                            (hasattr(self.read_mode, "value") and self.read_mode.value == "select")
                            or (self.read_mode == "select")
                        )

                )
            )
            else exclude.add("select_statement")
        )
        (
            include.add("schema_name")
            if (
                ((not self.select_statement))
                and (not self.call_statement)
                and (
                    (
                        (
                            self.read_mode
                            and (
                                (hasattr(self.read_mode, "value") and self.read_mode.value == "call")
                                or (self.read_mode == "call")
                            )
                        )
                    )
                    or (

                            self.read_mode
                            and (
                                (hasattr(self.read_mode, "value") and self.read_mode.value == "general")
                                or (self.read_mode == "general")
                            )

                    )
                )
            )
            else exclude.add("schema_name")
        )
        (
            include.add("table_name")
            if (
                ((not self.select_statement))
                and (not self.procedure_name)
                and (not self.call_statement)
                and (

                        self.read_mode
                        and (
                            (hasattr(self.read_mode, "value") and self.read_mode.value == "general")
                            or (self.read_mode == "general")
                        )

                )
            )
            else exclude.add("table_name")
        )
        (
            include.add("user_defined_function")
            if (

                    (
                        (
                            self.read_mode
                            and (
                                (hasattr(self.read_mode, "value") and self.read_mode.value == "call")
                                or (self.read_mode == "call")
                            )
                        )
                    )
                    or (

                            self.read_mode
                            and (
                                (hasattr(self.read_mode, "value") and self.read_mode.value == "call_statement")
                                or (self.read_mode == "call_statement")
                            )

                    )

            )
            else exclude.add("user_defined_function")
        )
        (
            include.add("add_proccode_column")
            if (
                ((not self.select_statement))
                and (not self.table_name)
                and (
                    (
                        (
                            self.read_mode
                            and (
                                (hasattr(self.read_mode, "value") and self.read_mode.value == "call")
                                or (self.read_mode == "call")
                            )
                        )
                    )
                    or (

                            self.read_mode
                            and (
                                (hasattr(self.read_mode, "value") and self.read_mode.value == "call_statement")
                                or (self.read_mode == "call_statement")
                            )

                    )
                )
            )
            else exclude.add("add_proccode_column")
        )
        (
            include.add("procedure_name")
            if (
                ((not self.call_statement))
                and (not self.select_statement)
                and (not self.table_name)
                and (

                        self.read_mode
                        and (
                            (hasattr(self.read_mode, "value") and self.read_mode.value == "call")
                            or (self.read_mode == "call")
                        )

                )
            )
            else exclude.add("procedure_name")
        )
        (
            include.add("select_statement_read_from_file_select")
            if (
                ((not self.schema_name))
                and (not self.table_name)
                and (

                        self.read_mode
                        and (
                            (hasattr(self.read_mode, "value") and self.read_mode.value == "select")
                            or (self.read_mode == "select")
                        )

                )
            )
            else exclude.add("select_statement_read_from_file_select")
        )
        (
            include.add("forward_row_data")
            if (
                ((not self.select_statement))
                and (not self.table_name)
                and (
                    (
                        (
                            self.read_mode
                            and (
                                (hasattr(self.read_mode, "value") and self.read_mode.value == "call")
                                or (self.read_mode == "call")
                            )
                        )
                    )
                    or (

                            self.read_mode
                            and (
                                (hasattr(self.read_mode, "value") and self.read_mode.value == "call_statement")
                                or (self.read_mode == "call_statement")
                            )

                    )
                )
            )
            else exclude.add("forward_row_data")
        )
        (
            include.add("schema_name")
            if (not self.select_statement)
            and (not self.call_statement)
            and (
                self.read_mode
                and (
                    (hasattr(self.read_mode, "value") and self.read_mode.value and "call" in str(self.read_mode.value))
                    or ("call" in str(self.read_mode))
                )
                and self.read_mode
                and (
                    (
                        hasattr(self.read_mode, "value")
                        and self.read_mode.value
                        and "general" in str(self.read_mode.value)
                    )
                    or ("general" in str(self.read_mode))
                )
            )
            else exclude.add("schema_name")
        )
        (
            include.add("procedure_name")
            if (not self.call_statement)
            and (not self.select_statement)
            and (not self.table_name)
            and (
                self.read_mode
                and (
                    (hasattr(self.read_mode, "value") and self.read_mode.value == "call") or (self.read_mode == "call")
                )
            )
            else exclude.add("procedure_name")
        )
        (
            include.add("select_statement_read_from_file_select")
            if (not self.schema_name)
            and (not self.table_name)
            and (
                self.read_mode
                and (
                    (hasattr(self.read_mode, "value") and self.read_mode.value == "select")
                    or (self.read_mode == "select")
                )
            )
            else exclude.add("select_statement_read_from_file_select")
        )
        (
            include.add("table_name")
            if (not self.select_statement)
            and (not self.procedure_name)
            and (not self.call_statement)
            and (
                self.read_mode
                and (
                    (hasattr(self.read_mode, "value") and self.read_mode.value == "general")
                    or (self.read_mode == "general")
                )
            )
            else exclude.add("table_name")
        )
        (
            include.add("select_statement")
            if (not self.schema_name)
            and (not self.table_name)
            and (not self.procedure_name)
            and (not self.call_statement)
            and (
                self.read_mode
                and (
                    (hasattr(self.read_mode, "value") and self.read_mode.value == "select")
                    or (self.read_mode == "select")
                )
            )
            else exclude.add("select_statement")
        )
        (
            include.add("add_proccode_column")
            if (not self.select_statement)
            and (not self.table_name)
            and (
                self.read_mode
                and (
                    (hasattr(self.read_mode, "value") and self.read_mode.value and "call" in str(self.read_mode.value))
                    or ("call" in str(self.read_mode))
                )
                and self.read_mode
                and (
                    (
                        hasattr(self.read_mode, "value")
                        and self.read_mode.value
                        and "call_statement" in str(self.read_mode.value)
                    )
                    or ("call_statement" in str(self.read_mode))
                )
            )
            else exclude.add("add_proccode_column")
        )
        (
            include.add("forward_row_data")
            if (not self.select_statement)
            and (not self.table_name)
            and (
                self.read_mode
                and (
                    (hasattr(self.read_mode, "value") and self.read_mode.value and "call" in str(self.read_mode.value))
                    or ("call" in str(self.read_mode))
                )
                and self.read_mode
                and (
                    (
                        hasattr(self.read_mode, "value")
                        and self.read_mode.value
                        and "call_statement" in str(self.read_mode.value)
                    )
                    or ("call_statement" in str(self.read_mode))
                )
            )
            else exclude.add("forward_row_data")
        )
        (
            include.add("row_limit")
            if (
                self.read_mode
                and (
                    (
                        hasattr(self.read_mode, "value")
                        and self.read_mode.value
                        and "call" not in str(self.read_mode.value)
                    )
                    or ("call" not in str(self.read_mode))
                )
                and self.read_mode
                and (
                    (
                        hasattr(self.read_mode, "value")
                        and self.read_mode.value
                        and "call_statement" not in str(self.read_mode.value)
                    )
                    or ("call_statement" not in str(self.read_mode))
                )
            )
            else exclude.add("row_limit")
        )
        (
            include.add("call_statement")
            if (not self.table_name)
            and (not self.select_statement)
            and (not self.procedure_name)
            and (
                self.read_mode
                and (
                    (hasattr(self.read_mode, "value") and self.read_mode.value == "call_statement")
                    or (self.read_mode == "call_statement")
                )
            )
            else exclude.add("call_statement")
        )
        (
            include.add("user_defined_function")
            if (
                self.read_mode
                and (
                    (hasattr(self.read_mode, "value") and self.read_mode.value and "call" in str(self.read_mode.value))
                    or ("call" in str(self.read_mode))
                )
                and self.read_mode
                and (
                    (
                        hasattr(self.read_mode, "value")
                        and self.read_mode.value
                        and "call_statement" in str(self.read_mode.value)
                    )
                    or ("call_statement" in str(self.read_mode))
                )
            )
            else exclude.add("user_defined_function")
        )
        return include, exclude

    def validate_target(self) -> tuple[set, set]:
        include = set()
        exclude = set()

        (
            include.add("create_statement")
            if (
                ((not self.static_statement))
                and (not self.procedure_name)
                and (not self.call_statement)
                and (
                    (
                        (
                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value != "call")
                                or (self.write_mode != "call")
                            )
                        )
                    )
                    and (

                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value != "call_statement")
                                or (self.write_mode != "call_statement")
                            )

                    )
                    and (

                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value != "static_statement")
                                or (self.write_mode != "static_statement")
                            )

                    )
                    and (

                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value != "update_statement")
                                or (self.write_mode != "update_statement")
                            )

                    )
                )
            )
            else exclude.add("create_statement")
        )
        (
            include.add("update_statement")
            if (
                ((not self.static_statement))
                and (
                    (
                        (
                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value == "update_statement")
                                or (self.write_mode == "update_statement")
                            )
                        )
                    )
                    or (

                            self.write_mode
                            and (
                                (
                                    hasattr(self.write_mode, "value")
                                    and self.write_mode.value == "update_statement_table_action"
                                )
                                or (self.write_mode == "update_statement_table_action")
                            )

                    )
                )
            )
            else exclude.add("update_statement")
        )
        (
            include.add("schema_name")
            if (
                ((not self.static_statement))
                and (not self.call_statement)
                and (
                    (
                        (
                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value != "call_statement")
                                or (self.write_mode != "call_statement")
                            )
                        )
                    )
                    and (

                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value != "static_statement")
                                or (self.write_mode != "static_statement")
                            )

                    )
                    and (

                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value != "update_statement")
                                or (self.write_mode != "update_statement")
                            )

                    )
                )
            )
            else exclude.add("schema_name")
        )
        (
            include.add("table_name")
            if (
                ((not self.static_statement))
                and (not self.call_statement)
                and (
                    (
                        (
                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value != "call")
                                or (self.write_mode != "call")
                            )
                        )
                    )
                    and (

                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value != "call_statement")
                                or (self.write_mode != "call_statement")
                            )

                    )
                    and (

                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value != "static_statement")
                                or (self.write_mode != "static_statement")
                            )

                    )
                    and (

                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value != "update_statement")
                                or (self.write_mode != "update_statement")
                            )

                    )
                )
            )
            else exclude.add("table_name")
        )
        (
            include.add("key_column_names")
            if (
                (((not self.update_statement)) and ((not self.static_statement)))
                and (
                    (
                        (
                            (
                                self.existing_table_action
                                and (
                                    (
                                        hasattr(self.existing_table_action, "value")
                                        and self.existing_table_action.value == "append"
                                    )
                                    or (self.existing_table_action == "append")
                                )
                            )
                        )
                        or (
                            (
                                self.existing_table_action
                                and (
                                    (
                                        hasattr(self.existing_table_action, "value")
                                        and self.existing_table_action.value == "merge"
                                    )
                                    or (self.existing_table_action == "merge")
                                )
                            )
                        )
                        or (
                            (
                                self.existing_table_action
                                and (
                                    (
                                        hasattr(self.existing_table_action, "value")
                                        and self.existing_table_action.value == "update"
                                    )
                                    or (self.existing_table_action == "update")
                                )
                            )
                        )
                    )
                    or (
                        (
                            (
                                self.write_mode
                                and (
                                    (hasattr(self.write_mode, "value") and self.write_mode.value == "insert")
                                    or (self.write_mode == "insert")
                                )
                            )
                        )
                        or (

                                self.write_mode
                                and (
                                    (hasattr(self.write_mode, "value") and self.write_mode.value == "merge")
                                    or (self.write_mode == "merge")
                                )

                        )
                        or (

                                self.write_mode
                                and (
                                    (hasattr(self.write_mode, "value") and self.write_mode.value == "update")
                                    or (self.write_mode == "update")
                                )

                        )
                    )
                )
                and (
                    (
                        (
                            (
                                self.existing_table_action
                                and (
                                    (
                                        hasattr(self.existing_table_action, "value")
                                        and self.existing_table_action.value == "append"
                                    )
                                    or (self.existing_table_action == "append")
                                )
                            )
                        )
                        or (
                            (
                                self.existing_table_action
                                and (
                                    (
                                        hasattr(self.existing_table_action, "value")
                                        and self.existing_table_action.value == "merge"
                                    )
                                    or (self.existing_table_action == "merge")
                                )
                            )
                        )
                        or (
                            (
                                self.existing_table_action
                                and (
                                    (
                                        hasattr(self.existing_table_action, "value")
                                        and self.existing_table_action.value == "update"
                                    )
                                    or (self.existing_table_action == "update")
                                )
                            )
                        )
                    )
                    or (
                        (
                            (
                                self.write_mode
                                and (
                                    (hasattr(self.write_mode, "value") and self.write_mode.value == "insert")
                                    or (self.write_mode == "insert")
                                )
                            )
                        )
                        or (

                                self.write_mode
                                and (
                                    (hasattr(self.write_mode, "value") and self.write_mode.value == "merge")
                                    or (self.write_mode == "merge")
                                )

                        )
                        or (

                                self.write_mode
                                and (
                                    (hasattr(self.write_mode, "value") and self.write_mode.value == "update")
                                    or (self.write_mode == "update")
                                )

                        )
                    )
                )
            )
            else exclude.add("key_column_names")
        )
        (
            include.add("table_action")
            if (
                (
                    (
                        (
                            (
                                self.write_mode
                                and (
                                    (hasattr(self.write_mode, "value") and self.write_mode.value == "insert")
                                    or (self.write_mode == "insert")
                                )
                            )
                        )
                        or (
                            (
                                self.write_mode
                                and (
                                    (hasattr(self.write_mode, "value") and self.write_mode.value == "merge")
                                    or (self.write_mode == "merge")
                                )
                            )
                        )
                        or (
                            (
                                self.write_mode
                                and (
                                    (
                                        hasattr(self.write_mode, "value")
                                        and self.write_mode.value == "update_statement_table_action"
                                    )
                                    or (self.write_mode == "update_statement_table_action")
                                )
                            )
                        )
                    )
                )
                and (
                    ((not self.static_statement))
                    and (
                        (
                            (
                                self.write_mode
                                and (
                                    (hasattr(self.write_mode, "value") and self.write_mode.value != "static_statement")
                                    or (self.write_mode != "static_statement")
                                )
                            )
                        )
                        and (

                                self.write_mode
                                and (
                                    (hasattr(self.write_mode, "value") and self.write_mode.value != "update_statement")
                                    or (self.write_mode != "update_statement")
                                )

                        )
                    )
                )
            )
            else exclude.add("table_action")
        )
        (
            include.add("static_statement")
            if (
                ((not self.schema_name))
                and (not self.table_name)
                and (not self.update_statement)
                and (not self.procedure_name)
                and (not self.call_statement)
                and (

                        self.write_mode
                        and (
                            (hasattr(self.write_mode, "value") and self.write_mode.value == "static_statement")
                            or (self.write_mode == "static_statement")
                        )

                )
            )
            else exclude.add("static_statement")
        )
        (
            include.add("call_statement")
            if (
                ((not self.static_statement))
                and (not self.schema_name)
                and (not self.table_name)
                and (not self.update_statement)
                and (not self.procedure_name)
                and (

                        self.write_mode
                        and (
                            (hasattr(self.write_mode, "value") and self.write_mode.value == "call_statement")
                            or (self.write_mode == "call_statement")
                        )

                )
            )
            else exclude.add("call_statement")
        )
        (
            include.add("update_statement_read_from_file_update")
            if (
                ((not self.static_statement))
                and (
                    (
                        (
                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value == "update_statement")
                                or (self.write_mode == "update_statement")
                            )
                        )
                    )
                    or (

                            self.write_mode
                            and (
                                (
                                    hasattr(self.write_mode, "value")
                                    and self.write_mode.value == "update_statement_table_action"
                                )
                                or (self.write_mode == "update_statement_table_action")
                            )

                    )
                )
            )
            else exclude.add("update_statement_read_from_file_update")
        )
        (
            include.add("procedure_name")
            if (
                ((not self.static_statement))
                and (not self.table_name)
                and (not self.update_statement)
                and (not self.call_statement)
                and (

                        self.write_mode
                        and (
                            (hasattr(self.write_mode, "value") and self.write_mode.value == "call")
                            or (self.write_mode == "call")
                        )

                )
            )
            else exclude.add("procedure_name")
        )
        (
            include.add("batch_size")
            if (((not self.has_reject_output)) or (not self.has_reject_output))
            else exclude.add("batch_size")
        )
        (
            include.add("call_each_row")
            if (
                ((not self.static_statement))
                and (not self.table_name)
                and (not self.update_statement)
                and (
                    (
                        (
                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value == "call")
                                or (self.write_mode == "call")
                            )
                        )
                    )
                    or (

                            self.write_mode
                            and (
                                (hasattr(self.write_mode, "value") and self.write_mode.value == "call_statement")
                                or (self.write_mode == "call_statement")
                            )

                    )
                )
            )
            else exclude.add("call_each_row")
        )
        (
            include.add("schema_name")
            if (not self.static_statement)
            and (not self.call_statement)
            and (
                self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "call_statement" not in str(self.write_mode.value)
                    )
                    or ("call_statement" not in str(self.write_mode))
                )
                and self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "static_statement" not in str(self.write_mode.value)
                    )
                    or ("static_statement" not in str(self.write_mode))
                )
                and self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "update_statement" not in str(self.write_mode.value)
                    )
                    or ("update_statement" not in str(self.write_mode))
                )
            )
            else exclude.add("schema_name")
        )
        (
            include.add("procedure_name")
            if (not self.static_statement)
            and (not self.table_name)
            and (not self.update_statement)
            and (not self.call_statement)
            and (
                self.write_mode
                and (
                    (hasattr(self.write_mode, "value") and self.write_mode.value == "call")
                    or (self.write_mode == "call")
                )
            )
            else exclude.add("procedure_name")
        )
        (
            include.add("key_column_names")
            if ((not self.update_statement) and (not self.static_statement))
            and (
                (
                    self.existing_table_action
                    and (
                        (
                            hasattr(self.existing_table_action, "value")
                            and self.existing_table_action.value
                            and "append" in str(self.existing_table_action.value)
                        )
                        or ("append" in str(self.existing_table_action))
                    )
                    or self.existing_table_action
                    and (
                        (
                            hasattr(self.existing_table_action, "value")
                            and self.existing_table_action.value
                            and "merge" in str(self.existing_table_action.value)
                        )
                        or ("merge" in str(self.existing_table_action))
                    )
                    or self.existing_table_action
                    and (
                        (
                            hasattr(self.existing_table_action, "value")
                            and self.existing_table_action.value
                            and "update" in str(self.existing_table_action.value)
                        )
                        or ("update" in str(self.existing_table_action))
                    )
                )
                or (
                    self.write_mode
                    and (
                        (
                            hasattr(self.write_mode, "value")
                            and self.write_mode.value
                            and "insert" in str(self.write_mode.value)
                        )
                        or ("insert" in str(self.write_mode))
                    )
                    or self.write_mode
                    and (
                        (
                            hasattr(self.write_mode, "value")
                            and self.write_mode.value
                            and "merge" in str(self.write_mode.value)
                        )
                        or ("merge" in str(self.write_mode))
                    )
                    or self.write_mode
                    and (
                        (
                            hasattr(self.write_mode, "value")
                            and self.write_mode.value
                            and "update" in str(self.write_mode.value)
                        )
                        or ("update" in str(self.write_mode))
                    )
                )
            )
            and (
                (
                    self.existing_table_action
                    and (
                        (
                            hasattr(self.existing_table_action, "value")
                            and self.existing_table_action.value
                            and "append" in str(self.existing_table_action.value)
                        )
                        or ("append" in str(self.existing_table_action))
                    )
                    or self.existing_table_action
                    and (
                        (
                            hasattr(self.existing_table_action, "value")
                            and self.existing_table_action.value
                            and "merge" in str(self.existing_table_action.value)
                        )
                        or ("merge" in str(self.existing_table_action))
                    )
                    or self.existing_table_action
                    and (
                        (
                            hasattr(self.existing_table_action, "value")
                            and self.existing_table_action.value
                            and "update" in str(self.existing_table_action.value)
                        )
                        or ("update" in str(self.existing_table_action))
                    )
                )
                or (
                    self.write_mode
                    and (
                        (
                            hasattr(self.write_mode, "value")
                            and self.write_mode.value
                            and "insert" in str(self.write_mode.value)
                        )
                        or ("insert" in str(self.write_mode))
                    )
                    or self.write_mode
                    and (
                        (
                            hasattr(self.write_mode, "value")
                            and self.write_mode.value
                            and "merge" in str(self.write_mode.value)
                        )
                        or ("merge" in str(self.write_mode))
                    )
                    or self.write_mode
                    and (
                        (
                            hasattr(self.write_mode, "value")
                            and self.write_mode.value
                            and "update" in str(self.write_mode.value)
                        )
                        or ("update" in str(self.write_mode))
                    )
                )
            )
            else exclude.add("key_column_names")
        )
        (
            include.add("create_statement")
            if (not self.static_statement)
            and (not self.procedure_name)
            and (not self.call_statement)
            and (
                self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "call" not in str(self.write_mode.value)
                    )
                    or ("call" not in str(self.write_mode))
                )
                and self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "call_statement" not in str(self.write_mode.value)
                    )
                    or ("call_statement" not in str(self.write_mode))
                )
                and self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "static_statement" not in str(self.write_mode.value)
                    )
                    or ("static_statement" not in str(self.write_mode))
                )
                and self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "update_statement" not in str(self.write_mode.value)
                    )
                    or ("update_statement" not in str(self.write_mode))
                )
            )
            else exclude.add("create_statement")
        )
        (
            include.add("static_statement")
            if (not self.schema_name)
            and (not self.table_name)
            and (not self.update_statement)
            and (not self.procedure_name)
            and (not self.call_statement)
            and (
                self.write_mode
                and (
                    (hasattr(self.write_mode, "value") and self.write_mode.value == "static_statement")
                    or (self.write_mode == "static_statement")
                )
            )
            else exclude.add("static_statement")
        )
        (
            include.add("table_name")
            if (not self.static_statement)
            and (not self.call_statement)
            and (
                self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "call" not in str(self.write_mode.value)
                    )
                    or ("call" not in str(self.write_mode))
                )
                and self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "call_statement" not in str(self.write_mode.value)
                    )
                    or ("call_statement" not in str(self.write_mode))
                )
                and self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "static_statement" not in str(self.write_mode.value)
                    )
                    or ("static_statement" not in str(self.write_mode))
                )
                and self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "update_statement" not in str(self.write_mode.value)
                    )
                    or ("update_statement" not in str(self.write_mode))
                )
            )
            else exclude.add("table_name")
        )
        (
            include.add("batch_size")
            if (not self.has_reject_output) or (self.has_reject_output != "true" or not self.has_reject_output)
            else exclude.add("batch_size")
        )
        (
            include.add("update_statement_read_from_file_update")
            if (not self.static_statement)
            and (
                self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "update_statement" in str(self.write_mode.value)
                    )
                    or ("update_statement" in str(self.write_mode))
                )
                and self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "update_statement_table_action" in str(self.write_mode.value)
                    )
                    or ("update_statement_table_action" in str(self.write_mode))
                )
            )
            else exclude.add("update_statement_read_from_file_update")
        )
        (
            include.add("update_statement")
            if (not self.static_statement)
            and (
                self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "update_statement" in str(self.write_mode.value)
                    )
                    or ("update_statement" in str(self.write_mode))
                )
                and self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "update_statement_table_action" in str(self.write_mode.value)
                    )
                    or ("update_statement_table_action" in str(self.write_mode))
                )
            )
            else exclude.add("update_statement")
        )
        (
            include.add("call_statement")
            if (not self.static_statement)
            and (not self.schema_name)
            and (not self.table_name)
            and (not self.update_statement)
            and (not self.procedure_name)
            and (
                self.write_mode
                and (
                    (hasattr(self.write_mode, "value") and self.write_mode.value == "call_statement")
                    or (self.write_mode == "call_statement")
                )
            )
            else exclude.add("call_statement")
        )
        (
            include.add("table_action")
            if (

                    self.write_mode
                    and (
                        (
                            hasattr(self.write_mode, "value")
                            and self.write_mode.value
                            and "insert" in str(self.write_mode.value)
                        )
                        or ("insert" in str(self.write_mode))
                    )
                    and self.write_mode
                    and (
                        (
                            hasattr(self.write_mode, "value")
                            and self.write_mode.value
                            and "merge" in str(self.write_mode.value)
                        )
                        or ("merge" in str(self.write_mode))
                    )
                    and self.write_mode
                    and (
                        (
                            hasattr(self.write_mode, "value")
                            and self.write_mode.value
                            and "update_statement_table_action" in str(self.write_mode.value)
                        )
                        or ("update_statement_table_action" in str(self.write_mode))
                    )

            )
            and (
                (not self.static_statement)
                and (
                    self.write_mode
                    and (
                        (
                            hasattr(self.write_mode, "value")
                            and self.write_mode.value
                            and "static_statement" not in str(self.write_mode.value)
                        )
                        or ("static_statement" not in str(self.write_mode))
                    )
                    and self.write_mode
                    and (
                        (
                            hasattr(self.write_mode, "value")
                            and self.write_mode.value
                            and "update_statement" not in str(self.write_mode.value)
                        )
                        or ("update_statement" not in str(self.write_mode))
                    )
                )
            )
            else exclude.add("table_action")
        )
        (
            include.add("call_each_row")
            if (not self.static_statement)
            and (not self.table_name)
            and (not self.update_statement)
            and (
                self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "call" in str(self.write_mode.value)
                    )
                    or ("call" in str(self.write_mode))
                )
                and self.write_mode
                and (
                    (
                        hasattr(self.write_mode, "value")
                        and self.write_mode.value
                        and "call_statement" in str(self.write_mode.value)
                    )
                    or ("call_statement" in str(self.write_mode))
                )
            )
            else exclude.add("call_each_row")
        )
        return include, exclude

    def get_source_props(self) -> dict:
        include, exclude = self.validate_source()
        props = {
            "add_proccode_column",
            "before_after_after_node_fail_on_error",
            "byte_limit",
            "call_statement",
            "default_max_string_binary_precision",
            "ds_java_heap_size",
            "enable_after_sql",
            "enable_after_sql_node",
            "enable_before_sql",
            "enable_before_sql_node",
            "fail_on_error_after_sql",
            "fail_on_error_before_sql",
            "fail_on_error_before_sql_node",
            "forward_row_data",
            "generate_unicode_columns",
            "has_ref_output",
            "key_column_names",
            "lookup_type",
            "procedure_name",
            "push_filters",
            "pushed_filters",
            "rcp",
            "read_after_sql_node_statements_from_file",
            "read_after_sql_statements_from_file",
            "read_before_sql_node_statement_from_file",
            "read_before_sql_statements_from_file",
            "read_mode",
            "rejected_filters",
            "row_limit",
            "sampling_percentage",
            "sampling_seed",
            "sampling_type",
            "schema_name",
            "select_statement",
            "select_statement_read_from_file_select",
            "table_name",
            "transform",
            "user_defined_function",
        }
        required = {"select_statement", "table_name"}
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_target_props(self) -> dict:
        include, exclude = self.validate_target()
        props = {
            "batch_size",
            "before_after_after_node_fail_on_error",
            "call_each_row",
            "call_statement",
            "create_statement",
            "ds_java_heap_size",
            "enable_after_sql",
            "enable_after_sql_node",
            "enable_before_sql",
            "enable_before_sql_node",
            "existing_table_action",
            "fail_on_error_after_sql",
            "fail_on_error_before_sql",
            "fail_on_error_before_sql_node",
            "has_reject_output",
            "key_column_names",
            "procedure_name",
            "read_after_sql_node_statements_from_file",
            "read_after_sql_statements_from_file",
            "read_before_sql_node_statement_from_file",
            "read_before_sql_statements_from_file",
            "schema_name",
            "static_statement",
            "table_action",
            "table_name",
            "update_statement",
            "update_statement_read_from_file_update",
            "write_mode",
        }
        required = {"static_statement"}
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_parameters_props(self) -> dict:
        include, exclude = self.validate_parameters()
        props = {"execmode", "input_count", "output_count", "preserve"}
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_app_data_props(self) -> dict:
        return {
            "datastage": {
                "active": 0,
                "SupportsRef": True,
                "maxRejectOutputs": 1,
                "minRejectOutputs": 0,
                "maxReferenceInputs": 0,
                "minReferenceInputs": 0,
            }
        }

    def get_input_ports_props(self) -> dict:
        include, exclude = self.validate_target()
        props = {"runtime_column_propagation"}
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property: {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties: {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_input_cardinality(self) -> dict:
        return {"min": 0, "max": 1}

    def get_output_ports_props(self) -> dict:
        include, exclude = self.validate_source()
        props = {
            "reject_condition_row_is_rejected",
            "reject_data_element_errorcode",
            "reject_data_element_errortext",
            "reject_from_link",
            "reject_number",
            "reject_threshold",
            "reject_uses",
        }
        required = set()
        props = props & (self.model_fields_set | required)
        conflict = exclude & self.model_fields_set & props
        if conflict:
            conflict = list(conflict)
            if len(conflict) == 1:
                warnings.warn(f"\n\033[33mFound conflicting property: {conflict[0]}\033[0m")
            else:
                warnings.warn(
                    f'\n\033[33mFound conflicting properties: {", ".join(conflict[:-1])} and {conflict[-1]}\033[0m'
                )
        exclude = exclude - self.model_fields_set
        return self.model_dump(include=props - exclude, by_alias=True, exclude_none=True, warnings=False)

    def get_output_cardinality(self) -> dict:
        return {"min": 0, "max": 1}

    def get_allowed_as_source_props(self) -> bool:
        return True

    def get_allowed_as_target_props(self) -> bool:
        return True



#  IBM Confidential
#  PID 5900-BAF
#  Copyright StreamSets Inc., an IBM Company 2025

"""Module containing Python Generator processors for streaming flow."""

import copy
import re
import textwrap
from collections import defaultdict
from ibm_watsonx_data_integration.codegen.code import Code, Coder
from ibm_watsonx_data_integration.codegen.preamble import StreamingPreamble
from ibm_watsonx_data_integration.common.constants import (
    PROD_BASE_API_URL,
    PROD_BASE_URL,
)
from ibm_watsonx_data_integration.services.streamsets.api import EnvironmentApiClient
from ibm_watsonx_data_integration.services.streamsets.models.flow_model import PipelineDefinition
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ibm_watsonx_data_integration.common.auth import IAMAuthenticator


class StageVertex:
    """Wrapper class for streaming flow stage."""

    def __init__(
        self, source_data: dict, config_definitions: list[dict[str, Any]], number_suffix: int | None = None
    ) -> None:
        """The __init__ of the StageVertex class.

        Args:
            source_data: Raw stage data taken from pipeline definition.
            config_definitions: Default stage configuration definitions.
            number_suffix: Number of stage in whole pipeline definition.
        """
        self._source_data = source_data
        self._config_definitions = config_definitions
        self._number_suffix = number_suffix

        # TODO: add event lanes
        self.output_lanes = source_data["outputLanes"]
        self.input_lanes = source_data["inputLanes"]
        # Remove additional stage numbers.
        self.label = re.sub(r"\s\d+$", "", source_data["uiInfo"]["label"])
        self.stage_type = source_data["uiInfo"]["stageType"]
        self.stage_name = source_data["stageName"]
        self.configuration: list[dict[str, Any]] = source_data["configuration"]

        self.stage_variable_name = self.label.lower().replace(" ", "_")
        if self._number_suffix:
            self.stage_variable_name = f"{self.stage_variable_name}_{self._number_suffix}"

    def __str__(self) -> str:
        """String representation of code to create stage."""
        stage_definition = f'{self.stage_variable_name} = flow.add_stage("{self.label}")'
        return f"{stage_definition}\n{self.stage_configuration()}".strip()

    def stage_configuration(self) -> str:
        """Return stage configuration properties changed by user."""
        changed_config_values = []
        for stage_config in self.configuration:
            if not stage_config.get("value"):
                continue

            value = stage_config.get("value")
            name = stage_config.get("name")
            config_definition = self._get_stage_config_definition(config_name=name)

            if value == config_definition["defaultValue"]:
                continue

            human_readable_config_name, _ = PipelineDefinition.get_attribute(config_definition)
            processed_value = value
            # TODO: WSDK-487
            if (
                config_definition.get("type", "").lower() in ("text", "string")
                or config_definition.get("fieldName", "").lower() == "connectionselection"
            ):
                processed_value = f'"{processed_value}"'

            if config_definition.get("type", "").lower() == "credential":
                processed_value = '"***"'

            changed_config_values.append(f"{self.stage_variable_name}.{human_readable_config_name} = {processed_value}")

        return "\n".join(changed_config_values)

    def _get_stage_config_definition(self, config_name: str) -> dict[str, Any]:
        for config in self._config_definitions:
            if config["name"] == config_name:
                return config


class StreamingProcessor(Coder):
    """Main processor for recreating script to create streaming flow.

    Responsible for creating ``StageVertex`` class foe each stage within pipeline definition.
    Handle building relation graph between stages using adjacency list.
    """

    def __init__(
        self,
        source_data: dict,
        auth: "IAMAuthenticator",  # pragma: allowlist secret
        base_url: str = PROD_BASE_URL,
        base_api_url: str = PROD_BASE_API_URL,
    ) -> None:
        """The __init__ of the StreamingProcessor class.

        Args:
            source_data: Streaming flow definition as python dictionary.
            auth: Authenticator instance.
            base_url: URL to IBM Cloud.
            base_api_url: URL to API endpoints.
        """
        self._source_data = source_data
        self._environment_api_client = EnvironmentApiClient(auth=auth, base_url=base_api_url)
        self._stage_counter = defaultdict(int)
        self._stages = list()
        self._preamble = StreamingPreamble(source_data=source_data, base_url=base_url, base_api_url=base_api_url)
        self._lanes_to_vertex_lookup = dict()
        self._graph = defaultdict(list)
        self._definitions = None

    @property
    def definitions(self) -> dict[str, list]:
        """Parsed stage definitions with configuration definition."""
        if not self._definitions:
            res_json = self._environment_api_client.get_library_definitions_for_engine_version(
                engine_version=self._source_data["entity"]["streamsets_flow"]["engine_version"]
            ).json()
            self._definitions = self._parse_definition_response(res_json)

        return self._definitions

    @staticmethod
    def _parse_definition_response(definition_json: dict) -> dict:
        result = dict()
        for _, stage_definition in definition_json["stageDefinitionMap"].items():
            stage_name = stage_definition["name"]
            result[stage_name] = copy.deepcopy(stage_definition["configDefinitions"])

        return result

    def _extract_stages(self) -> None:
        """Actual StaveVertex creation and building relation graph between stages."""
        for stage_dict in self._source_data["pipeline_definition"]["stages"]:
            self._stage_counter[stage_dict["stageName"]] += 1

            default_stage_definition = self.definitions[stage_dict["stageName"]]

            vertex = StageVertex(
                stage_dict,
                number_suffix=self._stage_counter[stage_dict["stageName"]],
                config_definitions=default_stage_definition,
            )
            self._stages.append(vertex)

            # build stages connection graph
            if vertex.stage_type.lower() == "source":
                self._lanes_to_vertex_lookup[vertex.output_lanes[0]] = vertex
                _ = self._graph[vertex.output_lanes[0]]
                continue

            # stages type other than `source`
            for output_lane in vertex.output_lanes:
                self._lanes_to_vertex_lookup[output_lane] = vertex

            for input_lane in vertex.input_lanes:
                self._graph[input_lane].append(vertex)

    def stages_as_str(self) -> str:
        """Collect string representation of all stages.

        Returns:
            A string representation of create stage code for all stages.
        """
        if len(self._stages) == 0:
            self._extract_stages()

        result = []
        for stage in self._stages:
            result.append(str(stage))

        return "\n".join(result)

    def stages_connection_as_str(self) -> str:
        """Collect string representation of connection between stages.

        Returns:
            A string representation of connection between stages.
        """
        if len(self._stages) == 0:
            self._extract_stages()

        result = []
        for output_lane, neighbours in self._graph.items():
            vertex = self._lanes_to_vertex_lookup[output_lane]

            for neighbour in neighbours:
                result.append(f"{vertex.stage_variable_name}.connect_output_to({neighbour.stage_variable_name})")

        return "\n".join(result)

    @property
    def script_footer(self) -> str:
        """Return generated script footer."""
        return "project.update_flow(flow)"

    def to_code(self) -> Code:
        """Returns object holding generated python script."""
        content = textwrap.dedent(f"""\
{self._preamble}

{self.stages_as_str()}

{self.stages_connection_as_str()}

{self.script_footer}
""")
        return Code(content=content)



# -------------------------------------------------------------------
# This file was auto-generated. Do not edit directly.
# Generated by: StageGeneratorV2
# -------------------------------------------------------------------

from typing import TYPE_CHECKING

from ibm_watsonx_data_integration.services.datastage.models.flow.dag import StageNode
from ibm_watsonx_data_integration.services.datastage.models.stage_models.java_integration import java_integration


if TYPE_CHECKING:
    from ibm_watsonx_data_integration.services.datastage.models.flow.dag import DAG


import os
# from ibm.datastage._framework.components.java_library import JavaLibrary


class JavaIntegrationStage(StageNode):
    def __init__(self, dag: "DAG", configuration: java_integration=None, *, label: str = None):
        super().__init__(dag, configuration=configuration, label=label)
        self.configuration: java_integration = configuration or java_integration()


    # def set_java_library(self, java_library: JavaLibrary):
    #    self.configuration.java_library = java_library
    #    self.configuration.asset_name = java_library.name
    #    self.configuration.classpath = (str(list(map(os.path.basename, [java_library.primary_file] + java_library.secondary_files))).replace(",", ";").replace("'", "").replace(" ", "")[1:-1])
